# coding: utf-8

import contextlib

from connection import Connection


class ConnectionPool(object):
    def __init__(self, uri, db_name, cache_size):
        """Connection Pool to a single mongo instance.

        :param uri: connection uri string
        :param db_name: mongo database name
        :param cache_size: (optional): maximum inactive cached connections for this pool
        """
        assert isinstance(db_name, (str, unicode))
        assert isinstance(cache_size, int)

        self._uri = uri
        self._cache_size = cache_size
        self._cache = []
        self._db_name = db_name

    @property
    def db_name(self):
        return self._db_name

    @contextlib.contextmanager
    def get_connection(self):
        if self._cache:
            conn = self._cache.pop(0)
        else:
            conn = Connection(self._uri)
            conn.simple_connect()
        try:
            yield conn
            if len(self._cache) < self._cache_size:
                self._cache.append(conn)
        except Exception:
            conn.close()
            raise

    def close(self):
        for i in self._cache:
            i.close()
        self._cache = []



# coding: utf-8

import unittest
import datetime

from bson import ObjectId

from mokito.fields import (
    make_field,
    IntField,
    FloatField,
    StringField,
    BooleanField,
    ObjectIdField,
    DateTimeField,
    ChoiceField
)
from mokito.errors import MokitoChoiceError


class TestFields(unittest.TestCase):

    def test_int_field(self):
        f = make_field(int)
        self.assertIsInstance(f, IntField)
        self.assertFalse(f.dirty)
        self.assertIsNone(f.get())
        self.assertIsNone(f.value)
        self.assertIsNone(f.self_value)
        f.set(123)
        self.assertEqual(123, f.get())
        self.assertEqual(123, f.value)
        self.assertEqual(123, f.self_value)
        f.set('456')
        self.assertEqual(456, f.get())
        self.assertEqual(456, f.value)
        self.assertEqual(456, f.self_value)
        f.value = 789
        self.assertEqual(789, f.get())
        self.assertEqual(789, f.value)
        self.assertEqual(789, f.self_value)

        with self.assertRaises(ValueError):
            f.set('aaa')
        self.assertEqual(789, f.value)

    def test_float_field(self):
        f = make_field(float)
        self.assertIsInstance(f, FloatField)
        self.assertIsNone(f.get())
        self.assertIsNone(f.value)
        self.assertIsNone(f.self_value)
        f.set(123)
        self.assertEqual(123.0, f.get())
        self.assertEqual(123.0, f.value)
        self.assertEqual(123.0, f.self_value)
        f.set('456.7')
        self.assertEqual(456.7, f.get())
        self.assertEqual(456.7, f.value)
        self.assertEqual(456.7, f.self_value)

    def test_str_field(self):
        f = make_field(str)
        self.assertIsInstance(f, StringField)
        self.assertIsNone(f.get())
        self.assertIsNone(f.value)
        self.assertIsNone(f.self_value)
        f.set(123)
        self.assertEqual('123', f.get())
        self.assertEqual('123', f.value)
        self.assertEqual('123', f.self_value)
        f.set('456.7')
        self.assertEqual('456.7', f.get())
        self.assertEqual('456.7', f.value)
        self.assertEqual('456.7', f.self_value)

    def test_bool_field(self):
        f = make_field(bool)
        self.assertIsInstance(f, BooleanField)
        self.assertIsNone(f.get())
        self.assertIsNone(f.value)
        self.assertIsNone(f.self_value)
        f.set(123)
        self.assertTrue(f.get())
        self.assertTrue(f.value)
        self.assertTrue(f.self_value)
        f.set(None)
        self.assertFalse(f.get())
        self.assertFalse(f.value)
        self.assertFalse(f.self_value)
        f.set(True)
        self.assertTrue(f.get())
        self.assertTrue(f.value)
        self.assertTrue(f.self_value)
        f.value = False
        self.assertFalse(f.get())
        self.assertFalse(f.value)
        self.assertFalse(f.self_value)

    def test_ObjectId_field(self):
        id1 = ObjectId()
        id2 = ObjectId()
        f = make_field(ObjectId)
        self.assertIsInstance(f, ObjectIdField)
        self.assertIsNone(f.get())
        self.assertIsNone(f.value)
        self.assertIsNone(f.self_value)
        self.assertIsNone(f.get(_format='json'))
        f.set(id1)
        self.assertEqual(id1, f.get())
        self.assertEqual(id1, f.value)
        self.assertEqual(id1, f.self_value)
        self.assertEqual(str(id1), f.get(_format='json'))
        f.value = str(id2)
        self.assertEqual(id2, f.get())
        self.assertEqual(id2, f.value)
        self.assertEqual(id2, f.self_value)
        self.assertEqual(str(id2), f.get(_format='json'))

    def test_datetime_field(self):
        dt1 = datetime.datetime.now()
        dt2 = datetime.datetime.utcnow()
        f = make_field(datetime.datetime)
        self.assertIsInstance(f, DateTimeField)
        self.assertIsNone(f.get())
        self.assertIsNone(f.value)
        self.assertIsNone(f.self_value)
        self.assertIsNone(f.get(as_json=True))
        f.set(dt1)
        self.assertEqual(dt1, f.get())
        self.assertEqual(dt1, f.value)
        self.assertEqual(dt1, f.self_value)
        self.assertEqual(dt1.replace(microsecond=0).isoformat(), f.get(_date_format='iso'))
        self.assertEqual(dt1.strftime('%d/%m/%y'), f.get(_date_format='%d/%m/%y'))
        f.set(dt2)
        self.assertEqual(dt2, f.get())
        self.assertEqual(dt2, f.value)
        self.assertEqual(dt2, f.self_value)
        self.assertEqual(dt2.replace(microsecond=0).isoformat(), f.get(_date_format='iso'))
        self.assertEqual(dt2.strftime('%d/%m/%y'), f.get(_date_format='%d/%m/%y'))
        f.set('14/12/16', _date_format='%d/%m/%y')
        self.assertEqual(datetime.datetime(2016, 12, 14), f.value)
        self.assertEqual(datetime.datetime(2016, 12, 14), f.self_value)

    def test_choice_field(self):
        f = make_field(ChoiceField({1: 'a', 2: 'b'}))

        with self.assertRaises(MokitoChoiceError):
            f.set(1)

        f.set(1, inner=True)
        self.assertEqual(f.value, 'a')
        self.assertEqual(f.get(), 'a')
        self.assertEqual(f.get(inner=True), 1)

        f.set('b')
        self.assertEqual(f.value, 'b')
        self.assertEqual(f.get(), 'b')
        self.assertEqual(f.get(inner=True), 2)

        f._val = 111
        self.assertIsNone(f.value)
        self.assertIsNone(f.get())



# coding: utf-8

from bson import SON, ObjectId

from tornado.gen import coroutine, Return

import message
from tools import fields2sort, norm_spec


class Cursor(object):
    """Cursor is a class used to call oeprations on a given db/collection using a specific
    connection pool. it will transparently release connections back to the pool after they receive
    responses
    """

    def __init__(self, collection_name, pool):
        assert isinstance(collection_name, basestring)
        assert isinstance(pool, object)

        self.__collection_name = collection_name
        self.__pool = pool

    @property
    def db_name(self):
        return self.__pool.db_name

    @property
    def collection_name(self):
        return self.__collection_name

    def _full_collection_name(self, collection=None):
        return u'%s.%s' % (self.db_name, collection or self.__collection_name)

    @property
    def full_collection_name(self):
        return self._full_collection_name()

    @coroutine
    def find_one(self, spec_or_id=None, fields=None, skip=0, sort=None, _is_command=False):
        """Get a single document from the database.
        """
        res = yield self.find(spec_or_id, fields, skip, limit=1,
                              sort=sort, _is_command=_is_command)
        raise Return(res[0] if res else None)

    @coroutine
    def find(self, spec=None, fields=None, skip=0, limit=0, snapshot=False, tailable=False,
             sort=None, max_scan=None, _is_command=False, hint=None, no_result=False):
        """Query the database.
        :param spec: (optional): _id or dict
        :param fields: (optional): a list of field names that should be returned in the result set,
            or a dict specifying the fields to return
        :param skip: (optional): the number of documents to omit (from the start of the result set)
            when returning the results
        :param limit: (optional): the maximum number of results to return
        :param snapshot: (optional): if True, snapshot mode will be used for this query. Snapshot
            mode assures no duplicates are returned, or objects missed, which were present at both
            the start and end of the query's execution. For details, see the snapshot documentation
            <http://dochub.mongodb.org/core/snapshot>
        :param tailable: (optional): the result of this find call will be a tailable cursor -
            tailable cursors aren't closed when the last data is retrieved but are kept open and
            the cursors location marks the final document's position. if more data is received
            iteration of the cursor will continue from the last document received. For details, see
            the tailable cursor documentation
            <https://docs.mongodb.com/manual/core/tailable-cursors/>
            <https://docs.mongodb.com/manual/reference/method/cursor.tailable/>
        :param sort: (optional):
            1) a list of (key, direction) pairs specifying the sort order for this query. See
               `pymongo.cursor.Cursor.sort` for details.
            2) a field name or list of field names with the sign of sorting direction ['f1', '-f2']
        :param max_scan: (optional): limit the number of documents examined when performing the
            query
        :param _is_command:
        :param hint: override MongoDB’s default index selection
            <https://docs.mongodb.com/manual/reference/method/cursor.hint/>
        :param no_result: without result
        """
        def query_options():
            options = 0
            if tailable:
                options |= message.QUERY_OPTIONS["tailable_cursor"]
            return options

        spec = norm_spec(spec)
        if not _is_command and "$query" not in spec:
            spec = SON({"$query": spec})

        if sort:
            sort = message.index_document(fields2sort(sort))
            if sort:
                spec["$orderby"] = sort

        if hint:
            spec["$hint"] = hint
        if snapshot:
            spec["$snapshot"] = True
        if max_scan:
            spec["$maxScan"] = max_scan
        if limit is None:
            limit = 0
        if skip is None:
            skip = 0
        if fields:
            if not isinstance(fields, dict):
                fields = dict((i, 1) for i in fields)

        col_name = self.full_collection_name
        data = None
        with self.__pool.get_connection() as conn:
            request = message.query(query_options(), col_name, skip, limit, spec, fields)
            res = yield conn.send_message(request)
            cursor_id = res['cursor_id']
            if not no_result:
                data = res['data']
                while res['number_returned'] > 100:
                    request = message.get_more(col_name, 101, cursor_id)
                    res = yield conn.send_message(request)
                    data += res['data']

            request = message.kill_cursors(cursor_id)
            yield conn.send_message(request, False)

        raise Return(None if no_result else data)

    @coroutine
    def find_and_modify(self, spec, update, fields=None, sort=None, upsert=False, new=False):
        """
        https://docs.mongodb.com/manual/reference/command/findAndModify/
        :param spec: None, _id or dict
        :param update: a dict or bson.son.SON instance specifying the document to be used for the
            update or (in the case of an upsert) insert
        :param fields: (optional) a list of field names that should be returned in the result set,
            or a dict specifying the fields to return
        :param sort:
        :param upsert: (optional) perform an upsert if True
        :param new: (optional) When true, returns the modified document rather than the original.
            The default is false.
        """
        command = SON([
            ("findAndModify", self.collection_name),
            ("query", norm_spec(spec)),
            ("new", new),
            ("update", update),
        ])

        if sort:
            sort = message.index_document(fields2sort(sort))
            if sort:
                command["sort"] = sort
        if upsert:
            command["upsert"] = upsert

        if fields:
            if not isinstance(fields, dict):
                fields = dict((i, 1) for i in fields)
            command['fields'] = fields

        with self.__pool.get_connection() as conn:
            request = message.query(0, self._full_collection_name("$cmd"), 0, 1, command)
            res = yield conn.send_message(request)
            cursor_id = res['cursor_id']
            request = message.kill_cursors(cursor_id)
            yield conn.send_message(request, False)

        data = res['data'][0]
        if int(data['ok']):
            raise Return(data['value'])

    @coroutine
    def insert(self, doc_or_docs, safe=True, check_keys=True, **kwargs):
        """Insert a document(s) into this collection.

        :param doc_or_docs: a document or list of documents to be inserted
        :param safe: (optional): check that the insert succeeded
        :param check_keys: (optional): check if keys start with '$' or contain '.',
            raising `pymongo.errors.InvalidName` in either case. If `safe` is `True` then the
            insert will be checked for errors, raising `pymongo.errors.OperationFailure` if one
            occurred. Safe inserts wait for a response from the database, while normal inserts do
            not.
        :param kwargs: (optional): any additional arguments imply `safe=True`, and will be used as
            options for the `getLastError` command
        """
        docs = doc_or_docs
        if isinstance(docs, dict):
            docs = [docs]

        for i in docs:
            if '_id' not in i:
                i['_id'] = ObjectId()
        _ids = [i['_id'] for i in docs]

        if kwargs:
            safe = True

        with self.__pool.get_connection() as conn:
            request = message.insert(self.full_collection_name, docs, check_keys, safe, kwargs)
            yield conn.send_message(request, safe)

        raise Return(_ids[0] if len(docs) == 1 else _ids)

    @coroutine
    def update(self, spec, update, upsert=False, safe=True, multi=False,
               no_replace=False, **kwargs):
        """Update a document(s) in this collection.

        :param spec: None, _id or dict
        :param update: a dict or bson.son.SON instance specifying the document to be used for the
            update or (in the case of an upsert) insert
        :param upsert: perform an upsert if True
        :param safe: (optional): check that the update succeeded
        :param multi: (optional): update all documents that match `spec`, rather than just the
            first matching document. The default value for `multi` is currently `False`, but this
            might eventually change to `True`. It is recommended that you specify this argument
            explicitly for all update operations in order to prepare your code for that change.
        :param no_replace: if `True` then {document} => {$set: {document}}
        :param kwargs: (optional): any additional arguments imply `safe=True`, and will be used as
            options for the `getLastError` command
        :raise TypeError: if either `spec` or `document` is not an instance of `dict`.
        :raise OperationFailure: If `safe` is `True` then the update will be checked for errors,
            raising `pymongo.errors.OperationFailure` if one occurred. Safe updates require a
            response from the database, while normal updates do not - thus, setting `safe` to
            `True` will negatively impact performance.

        There are many useful `update modifiers`_ which can be used when performing updates. For
        example, here we use the "$set" modifier to modify some fields in a matching document:
        Usage:
          >>> import mokito
          >>> db = mokito.Client('db_name')
          >>> db.test.insert({"x": "y", "a": "b"})
          ObjectId('...')
          >>> list(db.test.find())
          [{u'a': u'b', u'x': u'y', u'_id': ObjectId('...')}]
          >>> db.test.update({"x": "y"}, {"$set": {"a": "c"}})
          >>> list(db.test.find())
          [{u'a': u'c', u'x': u'y', u'_id': ObjectId('...')}]
        """
        if not isinstance(update, dict):
            raise TypeError("document must be an instance of dict")

        spec = norm_spec(spec)
        upsert = bool(upsert)
        safe = True if kwargs else bool(safe)
        if no_replace and not update.get("$set"):
            update = {"$set": update}

        if not update:
            raise Return(False)

        with self.__pool.get_connection() as conn:
            request = message.update(self.full_collection_name, upsert,
                                     multi, spec, update, safe, kwargs)
            res = yield conn.send_message(request, safe)

        if safe:
            data = res['data'][0]
            raise Return(data.get('updatedExisting', False) or data.get('upserted', False))

    @coroutine
    def remove(self, spec=None, safe=True, **kwargs):
        """
        Remove documents
        :param spec: (optional): _id or dict
        :param safe:
        :param kwargs:
        :return:
        """

        spec = norm_spec(spec)
        safe = True if kwargs else bool(safe)

        with self.__pool.get_connection() as conn:
            request = message.delete(self.full_collection_name, spec, safe, kwargs)
            res = yield conn.send_message(request, safe)

        if safe:
            raise Return(res['data'][0].get('n', 0))

    @coroutine
    def count(self, spec=None, hint=None):
        """
        Count documents
        :param spec: (optional): _id or dict
        :param hint:
        """
        command = SON({
            'count': self.__collection_name,
            "query": norm_spec(spec)
        })

        if hint:
            spec["$hint"] = hint

        with self.__pool.get_connection() as conn:
            request = message.query(0, self._full_collection_name('$cmd'), 0, -1, command)
            res = yield conn.send_message(request)

        raise Return(res['data'][0]['n'])

    @coroutine
    def distinct(self, key, spec=None):
        command = SON({
            "distinct": self.__collection_name,
            "key": key,
            "query": norm_spec(spec)
        })

        with self.__pool.get_connection() as conn:
            request = message.query(0, self._full_collection_name('$cmd'), 0, -1, command)
            res = yield conn.send_message(request)

        raise Return(res['data'][0]['values'])

    @coroutine
    def exists(self, spec=None):
        res = yield self.find(spec, ['_id'], limit=1)
        raise Return(bool(res))

    @coroutine
    def command(self, spec=None, safe=True):
        res = yield self.find(spec, limit=1, _is_command=True, no_result=not safe)
        raise Return(res[0] if res else None)



# coding: utf-8

import copy
import inspect
import datetime
try:
    import ujson as json
except ImportError:
    import json

import pytz
from bson import ObjectId
from dateutil.parser import parse

from errors import MokitoDBREFError, MokitoChoiceError
from tools import SEPARATOR


def make_field(rules, _parent=None):
    from orm import Model
    if rules is None:
        return NoneField(_parent=_parent)

    elif isinstance(rules, Field):
        x = copy.deepcopy(rules)
        x._parent = _parent
        return x

    else:
        if inspect.isclass(rules):
            _type = rules
            _val = None
        else:
            _type = type(rules)
            _val = rules

        if _type is int:
            return IntField(_default=_val, _parent=_parent)
        if _type is float:
            return FloatField(_default=_val, _parent=_parent)
        if _type is str or _type is unicode:
            return StringField(_default=_val, _parent=_parent)
        if _type is bool:
            return BooleanField(_default=_val, _parent=_parent)
        if _type is ObjectId:
            return ObjectIdField(_default=_val, _parent=_parent)
        if _type is datetime.datetime:
            return DateTimeField(_default=_val, _parent=_parent)

        if _type is list:
            return ListField(rules, _parent=_parent)
        if _type is tuple:
            return TupleField(rules, _parent=_parent)
        if _type is dict:
            return DictField(rules, _parent=_parent)

        if issubclass(_type, Model):
            return _type(_parent=_parent)

        if issubclass(_type, Field):
            x = _type(rules)
            x._parent = _parent
            return x
        raise TypeError()


class Field(object):

    def __init__(self, _default=None, _parent=None, **kwargs):
        self._val = None
        self._parent = _parent
        self._dirty = False
        self._default = _default
        if _default is not None:
            self.set(_default)

    def __str__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.value)

    @property
    def parent(self):
        return self._parent

    @property
    def dirty(self):
        return self._dirty

    def dirty_clear(self):
        self._dirty = False

    def _value(self):
        return self._val

    def set(self, value, **kwargs):
        res = self._val != value
        if res:
            self._val = value
            self._dirty = True

    value = property(_value, set)

    def get(self, **kwargs):
        return self._value()

    @property
    def self_value(self):
        return self._val

    def clear(self):
        self._val = self._default
        self._dirty = True


class UndefinedField(Field):
    def __getitem__(self, key):
        raise IndexError(key)

    def __setitem__(self, key, value):
        raise IndexError(key)

    def __delitem__(self, key):
        raise IndexError(key)

    def _value(self):
        return

    def set(self, value, **kwargs):
        pass

    value = property(_value, set)


class NoneField(Field):
    def __getitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        try:
            if isinstance(self._val, (list, tuple)):
                k1 = int(k1)
            ret = self._val[k1]
        except KeyError:
            ret = UndefinedField(_parent=self)
        if k2:
            ret = ret.__getitem__(k2)

        return ret if isinstance(ret, Field) else NoneField(ret)

    def set(self, value, **kwargs):
        if isinstance(value, tuple):
            value = list(value)
        super(NoneField, self).set(value, **kwargs)

    value = property(Field._value, set)


class NumberField(Field):
    def __iadd__(self, other):
        if isinstance(other, Field):
            other = other.value
        return (self._val or 0) + other


class IntField(NumberField):
    def set(self, value, **kwargs):
        if value is not None:
            value = int(value)
        super(IntField, self).set(value, **kwargs)

    value = property(NumberField._value, set)


class FloatField(NumberField):
    def set(self, value, **kwargs):
        if value is not None:
            value = float(value)
        super(FloatField, self).set(value, **kwargs)

    value = property(NumberField._value, set)


class StringField(Field):
    def __getitem__(self, key):
        return self._val[int(key)]

    def set(self, value, **kwargs):
        if value is not None:
            if isinstance(value, unicode):
                value = value.encode('utf-8')
            else:
                value = str(value)
        super(StringField, self).set(value, **kwargs)

    value = property(Field._value, set)


class BooleanField(Field):
    def set(self, value, **kwargs):
        if value is not None:
            value = bool(value)
        super(BooleanField, self).set(value, **kwargs)

    value = property(Field._value, set)


class ObjectIdField(Field):
    def get(self, _format=None, **kwargs):
        if self._val is not None:
            return str(self._val) if _format == 'json' else self._val

    def set(self, value, **kwargs):
        if value is not None:
            value = ObjectId(value)
        super(ObjectIdField, self).set(value, **kwargs)

    value = property(Field._value, set)


class DateTimeField(Field):
    def get(self, _date_format=None, tz_name=None, without_microsecond=True, tz=None, **kwargs):
        if _date_format is None and (tz_name or tz):
            _date_format = 'iso'
        if self._val is None or _date_format is None:
            return self._val

        _tz = pytz.timezone(tz_name) if tz_name else tz
        val = _tz.fromutc(self._val) if _tz else self._val
        if without_microsecond:
            val = val.replace(microsecond=0)

        return val.isoformat() if _date_format.lower() == 'iso' else val.strftime(_date_format)

    def set(self, value, _date_format=None, **kwargs):
        if not (value is None or isinstance(value, datetime.datetime)):
            if _date_format and _date_format != 'iso':
                value = datetime.datetime.strptime(value, _date_format)
            else:
                try:
                    value = parse(value).replace(tzinfo=None)
                except TypeError:
                    value = None
        super(DateTimeField, self).set(value, **kwargs)

    value = property(Field._value, set)


class ChoiceField(Field):
    def __init__(self, choices, **kwargs):
        """
        :param choices: {mongo_value: orm_value} or [mongo_value] or (mongo_value,)
        :param kwargs:
        """
        super(ChoiceField, self).__init__(**kwargs)
        if isinstance(choices, dict):
            self._choices = choices
        elif isinstance(choices, (list, tuple)):
            self._choices = {i: i for i in choices}
        else:
            raise TypeError()

    def _orm_2_mongo(self, value):
        for k, v in self._choices.items():
            if value == v:
                return k

        if value is not None:
            raise MokitoChoiceError(value)

    def get(self, inner=False, **kwargs):
        return self._val if inner else self._choices.get(self._val, None)

    def _value(self):
        return self._choices.get(self._val, None)

    def set(self, value, inner=False, **kwargs):
        if not inner:
            value = self._orm_2_mongo(value)
        super(ChoiceField, self).set(value, **kwargs)

    value = property(_value, set)


class CollectionField(Field):

    def __init__(self, _parent=None, **kwargs):
        super(CollectionField, self).__init__(_parent=_parent, **kwargs)
        self._add_docs = set()  # added Document()
        self._del_docs = set()  # deleted Document()

    def __getitem__(self, key):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError

    def __setitem__(self, key, value):
        self.setitem(key, value)

    def setitem(self, key, value, **kwargs):
        raise NotImplementedError

    def set(self, value, **kwargs):
        if value is None:
            self.clear()

        elif isinstance(value, dict):
            for k, v in value.items():
                self.setitem(k, v, **kwargs)

        else:
            raise TypeError()

    def dirty_clear(self):
        from orm import Document
        self._dirty = False
        self._add_docs = set()
        self._del_docs = set()
        for i in self._val.values():
            if not isinstance(i, Document):
                i.dirty_clear()

    @property
    def dirty(self):
        if not self._dirty:
            self._dirty = any(i.dirty for i in self._val.values())
        return self._dirty


class ArrayField(CollectionField):
    def __getitem__(self, key):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError

    def __len__(self):
        return max(map(int, self._val.keys())) + 1 if self._val else 0

    def __iter__(self):
        for i in range(len(self)):
            yield self.__getitem__(i)

    def setitem(self, key, value, **kwargs):
        raise NotImplementedError

    def get(self, key=None, **kwargs):
        if not (isinstance(key, (list, tuple)) or key is None):
            key = [key]
        return [self.__getitem__(i).get(**kwargs) for i in key or range(len(self))]

    def set(self, value, **kwargs):
        if isinstance(value, (list, tuple)):
            value = {str(k): v for k, v in enumerate(value)}
        super(ArrayField, self).set(value, **kwargs)

    def _value(self):
        return [self._val[i].value if i in self._val else None for i in map(str, range(len(self)))]

    value = property(_value, set)

    @property
    def self_value(self):
        return [self._val[i].self_value if i in self._val else None for i in map(str, range(len(self)))]

    def _set(self, k1, k2, value, **kwargs):
        from orm import Document, Model
        try:
            x = self._val[k1]
        except KeyError:
            raise IndexError(k1)
        if k2:
            x.setitem(k2, value, **kwargs)
        else:
            # if isinstance(value, Model):
            #     value = value.value

            if isinstance(x, Document):
                if value is None:
                    self._del_docs.add(k1)
                    self._add_docs.discard(k1)
                    del self._val[k1]
                elif isinstance(value, Document):
                    self._del_docs.discard(k1)
                    self._add_docs.add(k1)
                    self._val[k1] = value
                    if x.id != value.id:
                        self._dirty = True
                else:
                    self._del_docs.discard(k1)
                    self._add_docs.add(k1)
                    x.set(value, **kwargs)
            else:
                if isinstance(value, Model):
                    value = value.value

                x.set(value, **kwargs)

    @property
    def query(self):
        from orm import Document
        ret = {"$set": []}
        for k in range(len(self)):
            k = str(k)
            v = self._val.get(k)

            if k in self._add_docs:
                # if v._id is None or v.query:
                if getattr(v, '_id') is None or v.query:
                    raise MokitoDBREFError(v)
                ret["$set"].append(v.dbref)

            elif k in self._del_docs:
                ret["$set"].append(None)

            elif self.dirty:
                if v is None:
                    ret["$set"].append(None)

                elif isinstance(v, Document):
                    # if v._id is None or v.query:
                    if getattr(v, '_id') is None or v.query:
                        raise MokitoDBREFError(v)
                    ret["$set"].append(v.dbref)

                else:
                    # ret["$set"].append(v.value)
                    ret["$set"].append(v.get(inner=True))

        if not (ret["$set"] or self.dirty):
            del ret["$set"]
        return ret


class ListField(ArrayField):
    def __init__(self, rules, _parent=None, **kwargs):
        super(ListField, self).__init__(_parent=_parent, **kwargs)
        self._val = {}

        if not rules:
            rules = [None]

        # self._rules = make_field((rules+[None])[0], _parent=self)

        _rules = [make_field(i, _parent=self) for i in rules]
        # self._rules = _rules[0]
        self._rules = copy.deepcopy(_rules[0])
        for k, v in enumerate(_rules):
            if v._default:
                v._parent = self
                self._val[str(k)] = v

    def __getitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        try:
            ret = self._val[k1]
        except KeyError:
            ret = UndefinedField(_parent=self)
        return ret.__getitem__(k2) if k2 else ret

    def setitem(self, key, value, **kwargs):
        k1, _, k2 = str(key).partition(SEPARATOR)
        if not k1.isdigit():
            raise TypeError('%s is not integer' % k1)
        if k1 not in self._val:
            x = copy.deepcopy(self._rules)
            x._parent = self
            self._val[k1] = x
        self._set(k1, k2, value, **kwargs)

    def __delitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        try:
            x = self._val[k1]
        except KeyError:
            raise IndexError
        if k2:
            del x[k2]
            self._dirty = True
        else:
            self.pop(int(k1))

    def clear(self):
        self._val = {}
        self._dirty = True

    def append(self, value):
        self.setitem(len(self), value)
        # self._dirty = True

    def pop(self, key=-1):
        k1, _, k2 = str(key).partition(SEPARATOR)
        if k2:
            return self._val[k1].pop(k2)

        k1 = int(k1)
        if k1 < 0:
            k1 += len(self)

        ret = None
        keys = map(int, self._val.keys())
        keys.sort()
        for i in keys:
            if k1 == i:
                ret = self._val[str(i)]
                self._dirty = True
                del self._val[str(i)]
            elif k1 < i:
                self._val[str(i - 1)] = self._val[str(i)]
                self._dirty = True
                del self._val[str(i)]

        if ret is None:
            ret = UndefinedField(_parent=self)
        return ret


class TupleField(ArrayField):

    def __init__(self, rules, _parent=None, **kwargs):
        super(TupleField, self).__init__(_parent=_parent, **kwargs)
        self._val = {str(k): make_field(v, _parent=self) for k, v in enumerate(rules)}

    def __getitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        try:
            ret = self._val[k1]
            return ret.__getitem__(k2) if k2 else ret
        except (KeyError, AttributeError):
            raise IndexError(key)

    def __delitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        try:
            x = self._val[k1]
        except KeyError:
            raise IndexError
        if k2:
            x.__delitem__(k2)
        else:
            x.set(None)
        self._dirty = True

    def setitem(self, key, value, **kwargs):
        k1, _, k2 = str(key).partition(SEPARATOR)
        if not k1.isdigit():
            raise TypeError('%s is not integer' % k1)
        self._set(k1, k2, value, **kwargs)

    def clear(self):
        for i in self._val.values():
            i.clear()
        self._dirty = True


class DictField(CollectionField):

    def __init__(self, rules, _parent=None, **kwargs):
        super(DictField, self).__init__(_parent=_parent, **kwargs)
        self._rules = bool(rules)
        if rules:
            self._val = {k: make_field(v, _parent=self) for k, v in rules.items()}
        else:
            self._val = {}

    def __getitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        try:
            ret = self._val[k1]
        except KeyError:
            ret = UndefinedField(_parent=self)
        return ret.__getitem__(k2) if k2 else ret

    def setitem(self, key, value, **kwargs):
        k1, _, k2 = str(key).partition(SEPARATOR)
        if self._rules:
            self._set(k1, k2, value, **kwargs)
        else:
            self._val[k1] = make_field(value, _parent=self)
            # raise NotImplementedError

    def __delitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        if k2:
            del self._val[k1][k2]
        elif not self._rules:
            del self._val[k1]
        else:
            self._val[k1].clear()
        self._dirty = True

    def _set(self, k1, k2, value, **kwargs):
        from orm import Document, Model
        try:
            x = self._val[k1]
        except KeyError:
            return
        if k2:
            x.setitem(k2, value, **kwargs)
        else:
            # if isinstance(value, Model):
            #     value = value.value

            if isinstance(x, Document):
                if value is None:
                    self._del_docs.add(k1)
                    self._add_docs.discard(k1)
                    del self[k1]
                elif isinstance(value, Document):
                    self._del_docs.discard(k1)
                    self._add_docs.add(k1)
                    self._val[k1] = value
                    if x.id != value.id:
                        self._dirty = True
                else:
                    self._del_docs.discard(k1)
                    self._add_docs.add(k1)
                    x.set(value, **kwargs)
            else:
                if isinstance(value, Model):
                    value = value.value

                x.set(value, **kwargs)

    def get(self, key=None, **kwargs):
        if key is None:
            key = self._val.keys()
        if isinstance(key, (list, tuple)):
            return {i: self.__getitem__(i).get(**kwargs) for i in key}
        return self.__getitem__(key).get(**kwargs)

    def _value(self):
        return {k: v.value for k, v in self._val.items()}

    value = property(_value, CollectionField.set)

    @property
    def self_value(self):
        return {k: v.self_value for k, v in self._val.items()}

    def items(self):
        return self._val.items()

    def clear(self):
        if self._rules:
            for i in self._val.values():
                i.clear()
        else:
            self._val = {}
        self._dirty = True

    @property
    def query(self):
        from orm import Model, Document
        ret = {"$set": {}, "$unset": {}}

        if not self._rules:
            ret['$set'] = self.get(inner=True)

        else:
            for k, v in self._val.items():
                if k in self._add_docs:
                    # if v._id is None or v.query:
                    if getattr(v, '_id') is None or v.query:
                        raise MokitoDBREFError(v)
                    ret["$set"][k] = v.dbref

                elif k in self._del_docs:
                    ret["$unset"][k] = ''

                elif v.dirty and not isinstance(v, Document):
                    if isinstance(v, Model):
                        _q = v.query
                        for j in ["$set", "$unset"]:
                            if j in _q:
                                for k1, v1 in _q[j].items():
                                    ret[j]["%s.%s" % (k, k1)] = v1

                    elif isinstance(v, ArrayField):
                        _q = v.query
                        if "$set" in _q:
                            ret["$set"][k] = _q["$set"]

                    elif isinstance(v, DictField):
                        _q = v.query
                        for j in ["$set", "$unset"]:
                            if j in _q:
                                if v._rules:
                                    for k1, v1 in _q[j].items():
                                        ret[j]["%s.%s" % (k, k1)] = v1
                                else:
                                    ret[j][k] = _q[j]

                    else:
                        ret['$set'][k] = v.value

        if not ret["$set"]:
            del ret["$set"]
        if not ret["$unset"]:
            del ret["$unset"]
        return ret



# coding: utf-8

from tornado.ioloop import IOLoop
from tornado.gen import coroutine

import mokito

TEST_DB_NAME = 'test_mokito'
TEST_DB_URI = "mongodb://127.0.0.1:27017"
TEST_COLLECTION3 = 'tst3'


class TestClass3A(mokito.orm.Document):
    __uri__ = TEST_DB_URI
    __database__ = TEST_DB_NAME
    __collection__ = TEST_COLLECTION3
    scheme = {
        'f_0': None,
        'f_1': str,
    }

    def info(self):
        return '%s: %s' % (self.__class__.__name__, self.scheme)


class TestClass3B(TestClass3A):
    scheme = {
        'f_3': int,
    }


@coroutine
def main():
    print
    print 'ASYNC1', mokito.KnownClasses.data
    print 'ASYNC2', TestClass3B.scheme

IOLoop.current().run_sync(main)



# coding: utf-8

from tornado.gen import coroutine, Return

from bson.son import SON

from pool import ConnectionPool
from errors import MokitoParamError
from cursor import Cursor


class Client(object):

    def __init__(self, db_name, uri='mongodb://127.0.0.1:27017', cache_size=None):
        """
        Client connection to represent a remote database.

        Internally Client maintains a pool of connections that will live beyond the life of this
        object.

        :param db_name: mongo database name
        :param uri: connection uri string
        :param cache_size: maximum inactive cached connections for this pool
        :return a Client instance that wraps a pool.ConnectionPool

        Usage:
            >>> import mokito
            >>> db = mokito.Client('db_name')
            >>> yield db.collection_name.find({...})
        """
        self._pool = ConnectionPool(uri, db_name, cache_size or 10)
        self._db_name = db_name

    def __getattr__(self, name):
        """Get a collection by name."""
        return self.get_cursor(name)

    def __getitem__(self, name):
        """Get a collection by name."""
        return self.get_cursor(name)

    def get_cursor(self, collection_name):
        """Get a cursor to a collection by name.
        :param collection_name: the name of the collection
        :raise MokitoParamError on names with unallowable characters.
        """
        if not collection_name:
            raise MokitoParamError("collection names cannot be empty")
        if "$" in collection_name and \
                not (collection_name.startswith("oplog.$main") or
                     collection_name.startswith("$cmd")):
            raise MokitoParamError("collection names must not contain '$': %r" % collection_name)
        if collection_name.startswith(".") or collection_name.endswith("."):
            raise MokitoParamError('collection names must not start or end with ".": %r' %
                                   collection_name)
        if "\x00" in collection_name:
            raise MokitoParamError("collection names must not contain the null character")
        return Cursor(collection_name, self._pool)

    @coroutine
    def collection_names(self):
        res = yield self["system.namespaces"].find()
        names = [i['name'] for i in res if i['name'].count('.') == 1]
        strip = len(self._db_name) + 1
        raise Return([i[strip:] for i in names])

    @coroutine
    def command(self, command, value=1, **kwargs):
        """Issue a MongoDB command.

        Send command `command` to the database and return the response. If `command` is an instance
        of :class:`basestring` then the command {`command`: `value`} will be sent. Otherwise,
        `command` must be an instance of :class:`dict` and will be sent as is.

        Any additional keyword arguments will be added to the final command document before it is
        sent.

        :param command: document representing the command to be issued, or the name of the command
            (for simple commands only).
            .. note:: the order of keys in the `command` document is significant (the "verb" must
               come first), so commands which require multiple keys (e.g. `findandmodify`) should
               use an instance of :class:`~bson.son.SON` or a string and kwargs instead of a Python
               `dict`.
        :param value: (optional): value to use for the command verb when `command` is passed as a
            string
        :param kwargs: (optional): additional keyword arguments will be added to the command
          document before it is sent

        For example, a command like ``{buildinfo: 1}`` can be sent using:
        >>> import mokito
        >>> db = mokito.Client('db_name')
        >>> db.command("buildinfo")

        For a command where the value matters, like ``{collstats: collection_name}`` we can do:
        >>> db.command("collstats", 'collection_name')

        For commands that take additional arguments we can use kwargs. So
        ``{filemd5: object_id, root: file_root}`` becomes:
        >>> db.command("filemd5", object_id, root=file_root)
        """

        if isinstance(command, basestring):
            command = SON([(command, value)])
        command.update(kwargs)
        res = yield self.get_cursor("$cmd").command(command)
        raise Return(res)

    def close(self):
        self._pool.close()



# coding: utf-8

import copy

from bson import ObjectId
from bson.errors import InvalidId
from tornado.ioloop import IOLoop
from tornado.testing import AsyncTestCase, gen_test
from tornado.gen import coroutine

import mokito

import tests.util as u


class ORMTestCase(AsyncTestCase):
    def get_new_ioloop(self):
        return IOLoop.instance()

    @coroutine
    def _init_collection(self):
        yield self.db[u.TEST_COLLECTION1].insert([u.col1_data1, u.col1_data2])
        yield self.db[u.TEST_COLLECTION2].insert([u.col2_data1, u.col2_data2])

    @coroutine
    def _drop_collection(self):
        yield self.db.command('drop', u.TEST_COLLECTION2)
        yield self.db.command('drop', u.TEST_COLLECTION1)

    def setUp(self):
        super(ORMTestCase, self).setUp()
        self.setUpHelper()

    @gen_test
    def setUpHelper(self):
        self.db = mokito.ModelManager.databases[u.TEST_DB_NAME]
        yield self._drop_collection()
        yield self._init_collection()

    def tearDown(self):
        yield self.tearDownHelper()
        super(ORMTestCase, self).tearDown()

    @gen_test
    def tearDownHelper(self):
        yield self._drop_collection()

    def test_databases(self):
        self.assertIsInstance(mokito.ModelManager.databases[u.TEST_DB_NAME], mokito.Client)

    @gen_test
    def test_find_one_1(self):
        x = yield u.Document1.find_one(u.col1_id1)
        self.assertEqual(x._id, u.col1_id1)
        y = yield u.Document1.by_id(u.col1_id1)
        self.assertEqual(y._id, u.col1_id1)

        data = copy.deepcopy(u.col1_data1)
        data.pop('_id')
        data['x3'][1] = 1
        self.assertDictEqual(data, x.value)
        self.assertDictEqual(data, y.value)

        x = yield u.Document1.find_one(u.col1_id2)
        y = yield u.Document1.by_id(u.col1_id2)
        self.assertEqual(x._id, u.col1_id2)
        self.assertEqual(y._id, u.col1_id2)
        data = copy.deepcopy(u.col1_data2)
        data.pop('_id')
        data.pop('foo')
        data['x3'][1] = 2
        self.assertDictEqual(data, x.value)
        self.assertDictEqual(data, y.value)

    @gen_test
    def test_find_one_2(self):
        _id = ObjectId()
        x = yield u.Document1.find_one(_id)
        self.assertIsNone(x)

        with self.assertRaises(InvalidId):
            yield u.Document1.find_one('foo')

        with self.assertRaises(mokito.errors.MokitoORMError):
            yield u.Document1.by_id(_id)

    @gen_test
    def test_find_1(self):
        x = yield u.Document1.find()
        self.assertEqual(len(x), 2)

        data = copy.deepcopy(u.col1_data1)
        data.pop('_id')
        data['x3'][1] = 1
        self.assertDictEqual(data, x[0].value)

        data = copy.deepcopy(u.col1_data2)
        data.pop('_id')
        data.pop('foo')
        data['x3'][1] = 2
        self.assertDictEqual(data, x[1].value)

    @gen_test
    def test_count(self):
        self.assertEqual((yield u.Document1.count()), 2)

    @gen_test
    def test_save_1(self):
        x = yield u.Document1.find_one(u.col1_id2)
        self.assertFalse(x.dirty)
        x['x1'] = 3
        self.assertTrue(x.dirty)
        self.assertDictEqual(x.query, {'$set': {'x1': 3.0}})
        y = yield x.save()
        self.assertTrue(y)
        self.assertFalse(x.dirty)

        data = yield self.db[u.TEST_COLLECTION1].find_one(u.col1_id2, ['x1'])
        self.assertEqual(data['x1'], 3.0)
        self.assertEqual(x['x1'].value, 3.0)

        data = yield self.db[u.TEST_COLLECTION1].count()
        self.assertEqual(data, 2)

    @gen_test
    def test_save_2(self):
        data = copy.deepcopy(u.col1_data1)
        del data['_id']

        x = u.Document1(data, inner=True)
        self.assertIsNone(x._id)
        self.assertTrue(x.dirty)
        data['x3'][1] = 1
        self.assertDictEqual(x.value, data)

        y = yield x.save()
        self.assertTrue(y)
        self.assertIsNotNone(x._id)
        self.assertFalse(x.dirty)

        data = yield self.db[u.TEST_COLLECTION1].find_one(x._id)
        data.pop('_id')
        data['x3'][1] = 1
        self.assertDictEqual(x.value, data)
        self.assertEqual((yield self.db[u.TEST_COLLECTION1].count()), 3)

    @gen_test
    def test_save_3(self):
        x = u.Document2()

        self.assertIsNone(x.id)
        self.assertEqual((yield self.db[u.TEST_COLLECTION2].count()), 2)

        res = yield x.save()
        self.assertTrue(res)

        self.assertDictEqual(x.value, {
            'f1': [None, None],
            'f2': {'a': None, 'b': None},
            'm1': {'x2': [], 'x3': [None, None], 'x1': None, 'x4': {'a': None, 'b': None}},
            'm2': [],
            'd1': {'x2': [], 'x3': [None, None], 'x1': None, 'x4': {'a': None, 'b': None}},
            'd2': []
        })
        self.assertIsNotNone(x.id)
        self.assertEqual((yield self.db[u.TEST_COLLECTION2].count()), 3)

        res = yield self.db[u.TEST_COLLECTION2].find_one({'_id': {'$nin': [u.col2_id1, u.col2_id2]}})
        self.assertIsNotNone(res.pop('_id'))
        self.assertDictEqual(res, {
            'f1': [None, None],
            'f2': {'a': None, 'b': None},
            'm1': {'x2': [], 'x3': [None, None], 'x1': None, 'x4': {'a': None, 'b': None}},
            'm2': [],
            'd2': [],
            'd1': None
        })

    @gen_test
    def test_save_4(self):
        x = u.Document2()
        x['f1.1'] = 'foo'
        x['f2.a'] = 123

        self.assertIsNone(x.id)
        self.assertEqual((yield self.db[u.TEST_COLLECTION2].count()), 2)

        res = yield x.save()
        self.assertTrue(res)

        self.assertDictEqual(x.value, {
            'f1': [None, 'foo'],
            'f2': {'a': 123, 'b': None},
            'm1': {'x2': [], 'x3': [None, None], 'x1': None, 'x4': {'a': None, 'b': None}},
            'm2': [],
            'd1': {'x2': [], 'x3': [None, None], 'x1': None, 'x4': {'a': None, 'b': None}},
            'd2': []
        })
        self.assertIsNotNone(x.id)
        self.assertEqual((yield self.db[u.TEST_COLLECTION2].count()), 3)

        res = yield self.db[u.TEST_COLLECTION2].find_one({'_id': {'$nin': [u.col2_id1, u.col2_id2]}})
        self.assertIsNotNone(res.pop('_id'))
        self.assertDictEqual(res, {
            'f1': [None, 'foo'],
            'f2': {'a': 123, 'b': None},
            'm1': {'x2': [], 'x3': [None, None], 'x1': None, 'x4': {'a': None, 'b': None}},
            'm2': [],
            'd2': [],
            'd1': None
        })

    @gen_test
    def test_load_1(self):
        d = yield u.Document2.find_one(u.col2_id1)
        self.assertDictEqual(d['d1'].value, {'x2': [], 'x3': [None, None], 'x1': None, 'x4': {'a': None, 'b': None}})
        self.assertEqual(d['d1'].dbref, u.col1_dbref1)

        yield d['d1'].reread()
        self.assertDictEqual(d['d1'].value, {
            'x1': 0.5,
            'x2': [u.dt1, u.dt2, u.dt3],
            'x3': [123, 1],
            'x4': {'a': 1, 'b': 2}
        })

    @gen_test
    def test_load_2(self):
        d = yield u.Document2.find()

        self.assertDictEqual(d['0.d1'].value, {'x2': [], 'x3': [None, None], 'x1': None, 'x4': {'a': None, 'b': None}})
        self.assertEqual(d['0.d1'].dbref, u.col1_dbref1)
        yield d['0.d1'].reread()
        self.assertDictEqual(d['0.d1'].value, {
            'x1': 0.5,
            'x2': [u.dt1, u.dt2, u.dt3],
            'x3': [123, 1],
            'x4': {'a': 1, 'b': 2}
        })

    @gen_test
    def test_to_prop(self):
        x = yield u.Document1.find_one(u.col1_id1)
        self.assertEqual(x.prop1, x['x1'].value + 2)
        self.assertEqual(x.prop2, x['x1'].value * 2)
        x['x1'] = 10
        self.assertEqual(x.prop1, 12.0)
        self.assertEqual(x.prop2, 20.0)

    @gen_test
    def test_to_json_1(self):
        d = yield u.Document1.find_one(u.col1_id1)

        data = d.get(['id', 'x1', 'x2', 'x3', 'x4'], _date_format='iso', _format='json')
        self.assertDictEqual(data, {
            'id': str(u.col1_id1),
            'x1': 0.5,
            'x2': ["2016-01-02T03:04:05", "2016-02-03T04:05:06", "2016-03-04T05:06:07"],
            'x3': [123, 1],
            'x4': {"a": 1, "b": 2}
        })

        data = d.get(['x2', 'x3', 'x4'], _date_format='iso', tz_name='Asia/Novosibirsk')
        self.assertDictEqual(data, {
            'x2': ['2016-01-02T09:04:05+06:00', '2016-02-03T10:05:06+06:00', '2016-03-04T11:06:07+06:00'],
            'x3': [123, 1],
            'x4': {"a": 1, "b": 2}
        })

        data = d.get(['x1', 'prop1', 'prop2'])
        self.assertDictEqual(data, {
            'x1': 0.5,
            'prop1': 2.5,
            'prop2': 1.0
        })

    @gen_test
    def test_to_json_2(self):
        d = yield u.Document1.find()

        data = d.get(['id', 'x1', 'x2', 'x3', 'x4'], _date_format='iso', _format='json')
        self.assertListEqual(data, [
            {
                'id': str(u.col1_id1),
                'x1': 0.5,
                'x2': ['2016-01-02T03:04:05', '2016-02-03T04:05:06', '2016-03-04T05:06:07'],
                'x3': [123, 1],
                'x4': {'a': 1, 'b': 2}
            }, {
                'id': str(u.col1_id2),
                'x1': 1.5,
                'x2': ['2016-04-05T06:07:08', '2016-05-06T07:08:09', '2016-06-07T08:09:10'],
                'x3': [45601, 2],
                'x4': {'a': 3, 'b': 4}
            }
        ])

        data = d.get('x2', tz_name='Asia/Novosibirsk')
        self.assertListEqual(data, [
            ['2016-01-02T09:04:05+06:00', '2016-02-03T10:05:06+06:00', '2016-03-04T11:06:07+06:00'],
            ['2016-04-05T12:07:08+06:00', '2016-05-06T13:08:09+06:00', '2016-06-07T14:09:10+06:00']
        ])

        data = d.get(['x1', 'prop1', 'prop2'])
        self.assertListEqual(data, [
            {'x1': 0.5, 'prop1': 2.5, 'prop2': 1.0},
            {'x1': 1.5, 'prop1': 3.5, 'prop2': 3.0}
        ])

    @gen_test
    def test_to_json_3(self):
        d = yield u.Document1.find()
        data = d.get(['id', 'x1'])
        self.assertListEqual(data, [{'id': str(u.col1_id1), 'x1': 0.5},
                                    {'id': str(u.col1_id2), 'x1': 1.5}])

    @gen_test
    def test_from_json_1(self):
        d = u.Document2()
        self.assertFalse(d.dirty)
        d.set(u.col2_data2, inner=True)
        self.assertTrue(d.dirty)
        self.assertEqual(d['d1'].dbref, u.col1_dbref2)

        yield d['d1'].reread()
        yield [i.reread() for i in d['d2']]
        self.assertDictEqual(d.value, {
            'f1': [45602, 'foo3'],
            'f2': {'a': 11, 'b': 12},
            'm1': {'x2': [u.dt11, u.dt12], 'x3': [7, 3], 'x1': 0.2, 'x4': {'a': 8, 'b': 8}},
            'm2': [{'x2': [u.xt1], 'x3': [8, 1], 'x1': 0.2, 'x4': {'a': 9, 'b': 9}},
                   {'x2': [u.xt2, u.xt3], 'x3': [9, 2], 'x1': 0.3, 'x4': {'a': 10, 'b': 10}}],
            'd2': [{'x2': [u.dt4, u.dt5, u.dt6], 'x3': [45601, 2], 'x1': 1.5, 'x4': {'a': 3, 'b': 4}}],
            'd1': {'x2': [u.dt4, u.dt5, u.dt6], 'x3': [45601, 2], 'x1': 1.5, 'x4': {'a': 3, 'b': 4}}})

    @gen_test
    def test_distinct(self):
        data = yield u.Document2.distinct('m1.x1')
        self.assertListEqual(data, [0.1, 0.2])

    @gen_test
    def test_set_document(self):
        x1 = yield u.Document1.find_one(u.col1_id2)
        x2 = yield u.Document2.find_one(u.col2_id1)

        yield x2.reread('d1')
        self.assertDictEqual(x2['d1'].value, {
            'x1': 0.5,
            'x2': [u.dt1, u.dt2, u.dt3],
            'x3': [123, 1],
            'x4': {'a': 1, 'b': 2}
        })

        x2['d1'] = x1
        self.assertTrue(x2.dirty)
        self.assertDictEqual(x2.query, {'$set': {'d1': u.col1_dbref2}})

        yield x2.reread('d1')
        self.assertDictEqual(x2['d1'].value, {
            'x1': 1.5,
            'x2': [u.dt4, u.dt5, u.dt6],
            'x3': [45601, 2],
            'x4': {'a': 3, 'b': 4}
        })

    @gen_test
    def test_documents_1(self):
        x1 = yield u.Document1.find()
        self.assertEqual(len(x1), 2)
        for i in x1:
            self.assertIn(i._id, [u.col1_id1, u.col1_id2])

        yield x1.remove()

        x2 = yield u.Document1.find()
        self.assertEqual(len(x2), 0)

        yield x1.save()

        x2 = yield u.Document1.find()
        self.assertEqual(len(x2), 2)

        d1 = [i.value for i in x1]
        d1.sort()
        d2 = [i.value for i in x2]
        d2.sort()
        self.assertListEqual(d1, d2)



# coding: utf-8

import unittest
import datetime

from bson import ObjectId

from mokito.fields import (
    make_field,
    IntField,
    FloatField,
    StringField,
    BooleanField,
    ObjectIdField,
    DateTimeField,
    DictField,
    TupleField,
    ListField,
    ChoiceField,
    UndefinedField
)


class TestCollectionFields(unittest.TestCase):
    def test_list_field_1(self):
        f = make_field([int])
        self.assertIsInstance(f, ListField)
        self.assertIsInstance(f._rules, IntField)

        f = make_field([float])
        self.assertIsInstance(f, ListField)
        self.assertIsInstance(f._rules, FloatField)

        f = make_field([str])
        self.assertIsInstance(f, ListField)
        self.assertIsInstance(f._rules, StringField)

        f = make_field([bool])
        self.assertIsInstance(f, ListField)
        self.assertIsInstance(f._rules, BooleanField)

        f = make_field([datetime.datetime])
        self.assertIsInstance(f, ListField)
        self.assertIsInstance(f._rules, DateTimeField)

        f = make_field([ObjectId])
        self.assertIsInstance(f, ListField)
        self.assertIsInstance(f._rules, ObjectIdField)

    def test_list_field_2(self):
        f = make_field([str])
        for i in [0, '0', 'foo']:
            x = f[i]
            self.assertIsInstance(x, UndefinedField)
            self.assertIsNone(x.value)

        with self.assertRaises(IndexError):
            _ = f['0.0']

    def test_list_field_3(self):
        f = make_field([str])
        self.assertEqual(f.value, [])
        self.assertEqual(f.self_value, [])
        self.assertEqual(f.get(), [])

        f[1] = 'foo'
        self.assertEqual(f.value, [None, 'foo'])
        self.assertEqual(f.self_value, [None, 'foo'])
        self.assertEqual(f.get(), [None, 'foo'])
        self.assertTrue(f == f[1]._parent)

        f.append('bar')
        self.assertEqual(f.value, [None, 'foo', 'bar'])
        self.assertEqual(f.self_value, [None, 'foo', 'bar'])
        self.assertEqual(f.get(), [None, 'foo', 'bar'])
        self.assertTrue(f == f[2]._parent)

        f['3'] = 'x'
        self.assertEqual(f.value, [None, 'foo', 'bar', 'x'])
        self.assertEqual(f.self_value, [None, 'foo', 'bar', 'x'])
        self.assertEqual(f.get(), [None, 'foo', 'bar', 'x'])
        self.assertEqual(f.get([2, 1]), ['bar', 'foo'])
        self.assertTrue(f == f[3]._parent)

        self.assertIsNone(f.pop(0).get())
        self.assertEqual(f.value, ['foo', 'bar', 'x'])
        self.assertEqual(f.self_value, ['foo', 'bar', 'x'])
        self.assertEqual(f.get(), ['foo', 'bar', 'x'])

        self.assertEqual(f.pop(1).value, 'bar')
        self.assertEqual(f.value, ['foo', 'x'])
        self.assertEqual(f.self_value, ['foo', 'x'])
        self.assertEqual(f.get(), ['foo', 'x'])

        self.assertEqual(f.pop().get(), 'x')
        self.assertEqual(f.value, ['foo'])
        self.assertEqual(f.self_value, ['foo'])
        self.assertEqual(f.get(), ['foo'])

        self.assertEqual(f.pop(3).get(), None)
        self.assertEqual(f.value, ['foo'])
        self.assertEqual(f.self_value, ['foo'])
        self.assertEqual(f.get(), ['foo'])

        f[2] = 'zzz'
        self.assertEqual(f.value, ['foo', None, 'zzz'])
        self.assertEqual(f.self_value, ['foo', None, 'zzz'])
        self.assertEqual(f.get(), ['foo', None, 'zzz'])
        self.assertTrue(f == f[2]._parent)

        del f[0]
        self.assertEqual(f.value, [None, 'zzz'])
        self.assertEqual(f.self_value, [None, 'zzz'])
        self.assertEqual(f.get(), [None, 'zzz'])

        f.set(['foo', 'bar'])
        self.assertEqual(f.value, ['foo', 'bar'])
        self.assertEqual(f.self_value, ['foo', 'bar'])
        self.assertEqual(f.get(), ['foo', 'bar'])
        self.assertTrue(f == f[0]._parent)
        self.assertTrue(f == f[1]._parent)

        f[1].set(None)
        self.assertEqual(f.value, ['foo', None])
        self.assertEqual(f.self_value, ['foo', None])
        self.assertEqual(f.get(), ['foo', None])
        self.assertEqual(f.get([0, 1]), ['foo', None])
        self.assertEqual(f.get([1, 0]), [None, 'foo'])
        self.assertTrue(f == f[1]._parent)

        f.value = None
        self.assertEqual(f.value, [])
        self.assertEqual(f.self_value, [])
        self.assertEqual(f.get(), [])

    def test_list_field_4(self):
        dt1 = datetime.datetime.utcnow()
        dt2 = datetime.datetime(2016, 1, 2, 3, 4, 5)

        f = make_field([datetime.datetime])
        self.assertEqual(f.value, [])
        self.assertEqual(f.self_value, [])
        self.assertEqual(f.get(), [])

        f.set([dt1, dt2])
        self.assertEqual(f.value, [dt1, dt2])
        self.assertEqual(f.self_value, [dt1, dt2])
        self.assertEqual(f.get(), [dt1, dt2])
        self.assertEqual(f.get(_date_format='iso'),
                         [dt1.replace(microsecond=0).isoformat(),
                          dt2.replace(microsecond=0).isoformat()])
        self.assertEqual(f.get(_date_format='%d/%m/%y'),
                         [dt1.strftime('%d/%m/%y'), dt2.strftime('%d/%m/%y')])
        self.assertEqual(f.get([1], _date_format='%d/%m/%y'), [dt2.strftime('%d/%m/%y')])
        self.assertEqual(f.get(1, _date_format='%d/%m/%y'), [dt2.strftime('%d/%m/%y')])
        self.assertTrue(f == f[0]._parent)
        self.assertTrue(f == f[1]._parent)

    def test_tuple_field_1(self):
        f = make_field((int, str))
        self.assertIsInstance(f, TupleField)
        self.assertIsInstance(f._val['0'], IntField)
        self.assertTrue(f == f._val['0']._parent)
        self.assertIsInstance(f._val['1'], StringField)
        self.assertTrue(f == f._val['1']._parent)
        self.assertListEqual(f.value, [None, None])

        for i in [0, '0']:
            self.assertIsNone(f[i].value)
            self.assertIsNone(f[i].self_value)

        for i in ['0.0', 'a', 2]:
            with self.assertRaises(IndexError):
                _ = f[i]

    def test_tuple_field_4(self):
        f = make_field((datetime.datetime, str))

        self.assertEqual(f.value, [None, None])
        self.assertEqual(f.self_value, [None, None])
        self.assertEqual(f.get(), [None, None])
        self.assertEqual(f.get(_date_format='iso'), [None, None])

        f[1] = 123
        self.assertEqual(f.value, [None, '123'])
        self.assertEqual(f.self_value, [None, '123'])
        self.assertEqual(f.get(), [None, '123'])
        self.assertTrue(f == f[1]._parent)

        dt1 = datetime.datetime(2016, 1, 2, 3, 4, 5)
        f[0] = dt1
        self.assertEqual(f.value, [dt1, '123'])
        self.assertEqual(f.self_value, [dt1, '123'])
        self.assertEqual(f.get(), [dt1, '123'])
        self.assertEqual(f.get(_date_format='iso'), ['2016-01-02T03:04:05', '123'])
        self.assertTrue(f == f[0]._parent)

        f[1] = None
        self.assertEqual(f.value, [dt1, None])
        self.assertEqual(f.self_value, [dt1, None])
        self.assertEqual(f.get(), [dt1, None])
        self.assertTrue(f == f[1]._parent)

        del f[0]
        self.assertEqual(f.value, [None, None])
        self.assertEqual(f.self_value, [None, None])
        self.assertEqual(f.get(), [None, None])
        self.assertTrue(f == f[0]._parent)

        f.set([dt1, 456])
        self.assertEqual(f.value, [dt1, '456'])
        self.assertEqual(f.self_value, [dt1, '456'])
        self.assertEqual(f.get(), [dt1, '456'])
        self.assertEqual(f.get(_date_format='%d/%m/%y'), [dt1.strftime('%d/%m/%y'), '456'])
        self.assertTrue(f == f[0]._parent)
        self.assertTrue(f == f[1]._parent)

    def test_dict_field_1(self):
        dt1 = datetime.datetime(2016, 1, 2, 3, 4, 5)
        dt2 = datetime.datetime(2016, 2, 3, 4, 5, 6)
        id1 = ObjectId()
        data = {
            'd': {
                'd1': datetime.datetime,
                'd2': [str],
                'd3': (ObjectId, str)
            },
            'l': [{'l1': int, 'l2': str}],
            't': (int, {'t1': datetime.datetime, 't2': str}),
        }
        f = make_field(data)
        self.assertIsInstance(f, DictField)
        self.assertDictEqual(f.value, {
            'd': {'d1': None, 'd2': [], 'd3': [None, None]},
            'l': [],
            't': [None, {'t1': None, 't2': None}],
        })
        self.assertDictEqual(f.self_value, {
            'd': {'d1': None, 'd2': [], 'd3': [None, None]},
            'l': [],
            't': [None, {'t1': None, 't2': None}],
        })
        self.assertDictEqual(f.get(), {
            'd': {'d1': None, 'd2': [], 'd3': [None, None]},
            'l': [],
            't': [None, {'t1': None, 't2': None}],
        })
        self.assertDictEqual(f.get('d'), {'d1': None, 'd2': [], 'd3': [None, None]})
        self.assertTrue(f == f['d']._parent)
        self.assertTrue(f == f['d']['d1']._parent._parent)
        self.assertTrue(f == f['d']['d2']._parent._parent)
        self.assertTrue(f == f['d']['d3']._parent._parent)
        self.assertTrue(f == f['d']['d3'][0]._parent._parent._parent)
        self.assertTrue(f == f['d']['d3'][1]._parent._parent._parent)
        self.assertTrue(f == f['l']._parent)
        self.assertTrue(f == f['t']._parent)
        self.assertTrue(f == f['t'][0]._parent._parent)
        self.assertTrue(f == f['t'][1]._parent._parent)

        self.assertIsInstance(f._val['d'], DictField)
        self.assertIsInstance(f['d'], DictField)
        self.assertDictEqual(f['d'].value, {'d1': None, 'd2': [], 'd3': [None, None]})
        self.assertDictEqual(f['d'].self_value, {'d1': None, 'd2': [], 'd3': [None, None]})
        self.assertDictEqual(f['d'].get(), {'d1': None, 'd2': [], 'd3': [None, None]})

        self.assertIsInstance(f._val['l'], ListField)
        self.assertIsInstance(f['l'], ListField)
        self.assertEqual(f['l'].value, [])
        self.assertEqual(f['l'].self_value, [])
        self.assertEqual(f['l'].get(), [])

        self.assertIsInstance(f._val['t'], TupleField)
        self.assertIsInstance(f['t'], TupleField)
        self.assertEqual(f['t'].value, [None, {'t1': None, 't2': None}])
        self.assertEqual(f['t'].self_value, [None, {'t1': None, 't2': None}])
        self.assertEqual(f['t'].get(), [None, {'t1': None, 't2': None}])

        self.assertDictEqual(f.get(['l', 't']), {'l': [], 't': [None, {'t1': None, 't2': None}]})
        self.assertDictEqual(f.get(['d.d3', 'd.d1', 't.1.t1']),
                             {'t.1.t1': None, 'd.d3': [None, None], 'd.d1': None})
        self.assertEqual(f['d'].get(['d1', 'd2']), {'d1': None, 'd2': []})

        f.set({
            'd': {
                'd1': dt1,
                'd2': ['foo'],
                'd3': (id1, 'bar')
            },
            'l': [{'l1': 456, 'l2': 'baz'}],
            't': (789, {'t1': dt2, 't2': 'x'})
        })
        self.assertTrue(f == f['d']._parent)
        self.assertTrue(f == f['d']['d1']._parent._parent)
        self.assertTrue(f == f['d']['d2']._parent._parent)
        self.assertTrue(f == f['d']['d3']._parent._parent)
        self.assertTrue(f == f['d']['d3'][0]._parent._parent._parent)
        self.assertTrue(f == f['d']['d3'][1]._parent._parent._parent)
        self.assertTrue(f == f['l']._parent)
        self.assertTrue(f == f['l'][0]._parent._parent)
        self.assertTrue(f == f['l'][0]['l1']._parent._parent._parent)
        self.assertTrue(f == f['l'][0]['l2']._parent._parent._parent)
        self.assertTrue(f == f['t']._parent)
        self.assertTrue(f == f['t'][0]._parent._parent)
        self.assertTrue(f == f['t'][1]._parent._parent)
        self.assertTrue(f == f['t'][1]['t1']._parent._parent._parent)
        self.assertTrue(f == f['t'][1]['t2']._parent._parent._parent)
        self.assertDictEqual(f.value, {
            'd': {
                'd1': dt1,
                'd2': ['foo'],
                'd3': [id1, 'bar']
            },
            'l': [{'l1': 456, 'l2': 'baz'}],
            't': [789, {'t1': dt2, 't2': 'x'}]
        })
        self.assertDictEqual(f.self_value, {
            'd': {
                'd1': dt1,
                'd2': ['foo'],
                'd3': [id1, 'bar']
            },
            'l': [{'l1': 456, 'l2': 'baz'}],
            't': [789, {'t1': dt2, 't2': 'x'}]
        })
        self.assertDictEqual(f.get(), {
            'd': {
                'd1': dt1,
                'd2': ['foo'],
                'd3': [id1, 'bar']
            },
            'l': [{'l1': 456, 'l2': 'baz'}],
            't': [789, {'t1': dt2, 't2': 'x'}],
        })
        self.assertDictEqual(f.get(['d.d1', 'd.d3', 't.1'], _date_format='iso', _format='json'), {
            'd.d3': [str(id1), 'bar'],
            'd.d1': dt1.replace(microsecond=0).isoformat(),
            't.1': {'t2': 'x', 't1': dt2.replace(microsecond=0).isoformat()}
        })

        self.assertDictEqual(f['d'].value, {'d1': dt1, 'd2': ['foo'], 'd3': [id1, 'bar']})
        self.assertDictEqual(f['d'].self_value, {'d1': dt1, 'd2': ['foo'], 'd3': [id1, 'bar']})
        self.assertDictEqual(f['d'].get(), {'d1': dt1, 'd2': ['foo'], 'd3': [id1, 'bar']})

        self.assertEqual(f['l'].value, [{'l2': 'baz', 'l1': 456}])
        self.assertEqual(f['l'].self_value, [{'l2': 'baz', 'l1': 456}])
        self.assertEqual(f['l'].get(), [{'l2': 'baz', 'l1': 456}])

        self.assertEqual(f['t'].value, [789, {'t1': dt2, 't2': 'x'}])
        self.assertEqual(f['t'].self_value, [789, {'t1': dt2, 't2': 'x'}])
        self.assertEqual(f['t'].get(), [789, {'t1': dt2, 't2': 'x'}])
        self.assertEqual(f.get(['l', 't']),
                         {'l': [{'l1': 456, 'l2': 'baz'}],
                          't': [789, {'t1': dt2, 't2': 'x'}]})

        f['d.d1'] = dt2
        f['d.d2'].append('abc')
        f['d.d3.1'] = 'xyz'
        f['l'].append({'l1': 100, 'l2': 100})
        f['t']['1.t1'] = dt1
        self.assertDictEqual(f.value, {
            'd': {
                'd1': dt2,
                'd2': ['foo', 'abc'],
                'd3': [id1, 'xyz']
            },
            'l': [{'l1': 456, 'l2': 'baz'}, {'l1': 100, 'l2': '100'}],
            't': [789, {'t1': dt1, 't2': 'x'}],
        })
        self.assertDictEqual(f.self_value, {
            'd': {
                'd1': dt2,
                'd2': ['foo', 'abc'],
                'd3': [id1, 'xyz']
            },
            'l': [{'l1': 456, 'l2': 'baz'}, {'l1': 100, 'l2': '100'}],
            't': [789, {'t1': dt1, 't2': 'x'}],
        })

        f.clear()
        self.assertDictEqual(f.value, {
            'd': {'d1': None, 'd2': [], 'd3': [None, None]},
            'l': [],
            't': [None, {'t1': None, 't2': None}],
        })

        f.set({
            'd.d1': dt1.isoformat(),
            'd.d2.1': 'foo',
            'd.d2.2': 'bar',
            'd.d2.0': 'baz',
            'd.d3.1': 'xxx',
            'd.d3.0': str(id1),
            'l.1.l1': 456,
            'l.1.l2': 'baz2',
            'l.0.l1': 123,
            'l.0.l2': 'baz1',
            't.1.t1': dt2.isoformat(),
            't.0': 789
        })
        self.assertDictEqual(f.value, {
            'd': {
                'd1': dt1,
                'd2': ['baz', 'foo', 'bar'],
                'd3': [id1, 'xxx']
            },
            'l': [{'l1': 123, 'l2': 'baz1'}, {'l1': 456, 'l2': 'baz2'}],
            't': [789, {'t1': dt2, 't2': None}],
        })

        del f['d.d1']
        self.assertDictEqual(f.value, {
            'd': {
                'd1': None,
                'd2': ['baz', 'foo', 'bar'],
                'd3': [id1, 'xxx']
            },
            'l': [{'l1': 123, 'l2': 'baz1'}, {'l1': 456, 'l2': 'baz2'}],
            't': [789, {'t1': dt2, 't2': None}],
        })

        del f['d']
        self.assertDictEqual(f.value, {
            'd': {
                'd1': None,
                'd2': [],
                'd3': [None, None]
            },
            'l': [{'l1': 123, 'l2': 'baz1'}, {'l1': 456, 'l2': 'baz2'}],
            't': [789, {'t1': dt2, 't2': None}],
        })

        del f['l.1.l2']
        self.assertDictEqual(f.value, {
            'd': {
                'd1': None,
                'd2': [],
                'd3': [None, None]
            },
            'l': [{'l1': 123, 'l2': 'baz1'}, {'l1': 456, 'l2': None}],
            't': [789, {'t1': dt2, 't2': None}],
        })

        del f['l.1']
        self.assertDictEqual(f.value, {
            'd': {
                'd1': None,
                'd2': [],
                'd3': [None, None]
            },
            'l': [{'l1': 123, 'l2': 'baz1'}],
            't': [789, {'t1': dt2, 't2': None}],
        })

        del f['l']
        self.assertDictEqual(f.value, {
            'd': {
                'd1': None,
                'd2': [],
                'd3': [None, None]
            },
            'l': [],
            't': [789, {'t1': dt2, 't2': None}],
        })

        del f['t.1']
        self.assertDictEqual(f.value, {
            'd': {
                'd1': None,
                'd2': [],
                'd3': [None, None]
            },
            'l': [],
            't': [789, {'t1': None, 't2': None}],
        })

        del f['t.0']
        self.assertDictEqual(f.value, {
            'd': {
                'd1': None,
                'd2': [],
                'd3': [None, None]
            },
            'l': [],
            't': [None, {'t1': None, 't2': None}],
        })

        del f['t.1.t1']
        self.assertDictEqual(f.value, {
            'd': {
                'd1': None,
                'd2': [],
                'd3': [None, None]
            },
            'l': [],
            't': [None, {'t1': None, 't2': None}],
        })

        with self.assertRaises(IndexError):
            del f['t.2']

        with self.assertRaises(KeyError):
            del f['foo']

        self.assertDictEqual(f.get(['foo']), {'foo': None})
        self.assertIsNone(f.get('foo'))

    def test_dict_field_2(self):
        ch1 = ['a', 'b']
        ch2 = {1: 'x', 2: 'y'}
        data = {
            's': ChoiceField(ch1),
            'l': [ChoiceField(ch2)],
            't': (int, ChoiceField(ch2)),
        }
        f = make_field(data)
        f['s'] = 'a'
        f['l'].append('x')
        f['t'] = [123, 'y']

        self.assertDictEqual(f.value, {'s': 'a', 'l': ['x'], 't': [123, 'y']})
        self.assertDictEqual(f.get(), {'s': 'a', 'l': ['x'], 't': [123, 'y']})
        self.assertDictEqual(f.get(inner=True), {'s': 'a', 'l': [1], 't': [123, 2]})
        self.assertDictEqual(f.self_value, {'s': 'a', 'l': [1], 't': [123, 2]})

    def test_list_query_1(self):
        f = make_field([str])
        self.assertDictEqual(f.query, {})
        f.append('foo')
        self.assertDictEqual(f.query, {'$set': ['foo']})
        f.append('bar')
        self.assertDictEqual(f.query, {'$set': ['foo', 'bar']})
        del f[0]
        self.assertDictEqual(f.query, {'$set': ['bar']})
        f.dirty_clear()
        self.assertDictEqual(f.query, {})
        f[1] = 'x'
        self.assertDictEqual(f.query, {'$set': ['bar', 'x']})
        f.pop()
        self.assertDictEqual(f.query, {'$set': ['bar']})
        f.dirty_clear()
        del f[0]
        self.assertDictEqual(f.query, {'$set': []})

    def test_list_query_2(self):
        f = make_field([{'a': int, 'b': str}])
        self.assertDictEqual(f.query, {})
        f.append({'a': 1})
        self.assertDictEqual(f.query, {'$set': [{'a': 1, 'b': None}]})
        f.append({'b': 'foo'})
        self.assertDictEqual(f.query, {'$set': [{'a': 1, 'b': None}, {'a': None, 'b': 'foo'}]})
        f[0] = {'a': 123, 'b': 'foo'}
        self.assertDictEqual(f.query, {'$set': [{'a': 123, 'b': 'foo'}, {'a': None, 'b': 'foo'}]})

    def test_list_query_3(self):
        f = make_field([{}])
        self.assertDictEqual(f.query, {})
        f.append({'a': 1})
        self.assertDictEqual(f.query, {'$set': [{'a': 1}]})
        f.append({'b': 'foo'})
        self.assertDictEqual(f.query, {'$set': [{'a': 1}, {'b': 'foo'}]})
        f[0] = {'x': 123, 'y': 'foo'}
        self.assertDictEqual(f.query, {'$set': [{'a': 1, 'y': 'foo', 'x': 123}, {'b': 'foo'}]})

    def test_tuple_query_1(self):
        f = make_field((int, str))
        self.assertDictEqual(f.query, {})
        f[0] = 123
        self.assertDictEqual(f.query, {'$set': [123, None]})
        f[1] = 'foo'
        self.assertDictEqual(f.query, {'$set': [123, 'foo']})
        f.dirty_clear()
        self.assertDictEqual(f.query, {})
        del f[0]
        self.assertDictEqual(f.query, {'$set': [None, 'foo']})

    def test_tuple_query_2(self):
        f = make_field((str, [int]))
        self.assertDictEqual(f.query, {})
        f.set(['foo', [1]])
        self.assertDictEqual(f.query, {'$set': ['foo', [1]]})
        f['1.2'] = 123
        self.assertDictEqual(f.query, {'$set': ['foo', [1, None, 123]]})

    def test_tuple_query_3(self):
        f = make_field((int, {'a': int}))
        self.assertDictEqual(f.query, {})
        f['1.a'] = 123
        self.assertDictEqual(f.query, {'$set': [None, {'a': 123}]})
        f[0] = 123
        self.assertDictEqual(f.query, {'$set': [123, {'a': 123}]})

    def test_dict_query_1(self):
        data = {
            'd1': int,
            'd2': [str],
            'd3': (int, str),
            'd4': {
                'dd1': int,
                'dd2': [str],
                'dd3': (int, str)
            },
            'd5': {}
        }
        f = make_field(data)
        self.assertDictEqual(f.query, {})
        f['d1'] = 123
        self.assertDictEqual(f.query, {'$set': {'d1': 123}})
        f['d2.0'] = 123
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['123']
        }})
        f['d2'][1] = 'foo'
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['123', 'foo']
        }})
        f['d2'].append('bar')
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['123', 'foo', 'bar']
        }})
        f['d2'].set(['z', 123])
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar']
        }})
        f['d3.0'] = 123
        f['d3.1'] = 123
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [123, '123']
        }})
        f['d3'] = [None, 'foo']
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo']
        }})
        f['d4.dd1'] = 456
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo'],
            'd4.dd1': 456
        }})
        f['d4.dd2'].append('bar')
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo'],
            'd4.dd1': 456,
            'd4.dd2': ['bar']
        }})
        f['d4.dd2.2'] = 'x'
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo'],
            'd4.dd1': 456,
            'd4.dd2': ['bar', None, 'x']
        }})
        f['d4.dd3.1'] = 456
        self.assertDictEqual(f.query, {'$set': {
            'd1': 123,
            'd2': ['z', '123', 'bar'],
            'd3': [None, 'foo'],
            'd4.dd1': 456,
            'd4.dd2': ['bar', None, 'x'],
            'd4.dd3': [None, '456']
        }})
        f.dirty_clear()
        self.assertDictEqual(f.query, {})
        f.set({
            'd1': 456,
            'd2': ['foo', 'bar'],
            'd3': [123, 'x'],
            'd4': {
                'dd1': 123,
                'dd2': ['a', 'b', 'c'],
                'dd3': [None, None]
            }
        })
        self.assertDictEqual(f.query, {'$set': {
            'd1': 456,
            'd2': ['foo', 'bar', 'bar'],
            'd3': [123, 'x'],
            'd4.dd1': 123,
            'd4.dd2': ['a', 'b', 'c'],
            'd4.dd3': [None, None]
        }})
        f['d5'] = {'a': 1, 'b': 'foo'}
        self.assertDictEqual(f.query, {'$set': {
            'd1': 456,
            'd2': ['foo', 'bar', 'bar'],
            'd3': [123, 'x'],
            'd4.dd1': 123,
            'd4.dd2': ['a', 'b', 'c'],
            'd4.dd3': [None, None],
            'd5': {'a': 1, 'b': 'foo'}
        }})

        f['d4.dd3.1'] = 123
        self.assertDictEqual(f.query, {'$set': {
            'd1': 456,
            'd2': ['foo', 'bar', 'bar'],
            'd3': [123, 'x'],
            'd4.dd1': 123,
            'd4.dd2': ['a', 'b', 'c'],
            'd4.dd3': [None, '123'],
            'd5': {'a': 1, 'b': 'foo'}
        }})

    def test_dict_query_2(self):
        f = make_field({})
        self.assertDictEqual(f.query, {})
        f.set({
            'd1': 456,
            'd2': ['foo', 'bar'],
            'd3': [123],
            'd4': {
                'dd1': 123,
                'dd2': ['a', 'b', 'c'],
            }
        })
        self.assertDictEqual(f.query, {'$set': {
            'd1': 456,
            'd2': ['foo', 'bar'],
            'd3': [123],
            'd4': {
                'dd1': 123,
                'dd2': ['a', 'b', 'c'],
            }
        }})



# coding: utf-8

from .client import Client
from .orm import Model, Document, Documents
from .manage import ModelManager
from .fields import ChoiceField as Choice
from .fields import UndefinedField as Undefined
import errors

__all__ = [
    "Client",
    "Document",
    "Documents",
    "Model",
    "ModelManager",
    "errors",
    "Choice",
    "Undefined"
]

version = "0.2.15"



"""Tools for creating `messages
<http://www.mongodb.org/display/DOCS/Mongo+Wire+Protocol>
"""

import random
import struct

from bson import BSON, _make_c_string, decode_all, DEFAULT_CODEC_OPTIONS, SON

from errors import (
    MokitoEmptyBulkError,
    MokitoDriverError,
    MokitoMasterChangedError,
    MokitoInvalidCursorError
)

MAX_INT32 = 2147483647
MIN_INT32 = -2147483648

ASCENDING = 1
"""Ascending sort order."""
DESCENDING = -1
"""Descending sort order."""

GEO2D = "2d"
"""Index specifier for a 2-dimensional `geospatial index`_.

.. _geospatial index: http://docs.mongodb.org/manual/core/2d/
"""


def _randint():
    """Generate a pseudo random 32 bit integer."""
    return random.randint(MIN_INT32, MAX_INT32)


_ZERO_32 = b'\x00\x00\x00\x00'


def __last_error(args):
    """Data to send to do a lastError."""
    cmd = SON([("getlasterror", 1)])
    cmd.update(args)
    return query(0, "admin.$cmd", 0, -1, cmd)

QUERY_OPTIONS = {
    "tailable_cursor": 2,
    "slave_okay": 4,
    "oplog_replay": 8,
    "no_timeout": 16
}

# OP_REPLY = 1  # Reply to a client request. responseTo is set.
# OP_MSG = 1000  # Generic msg command followed by a string.
OP_UPDATE = 2001  # Update document.
OP_INSERT = 2002  # Insert new document.
OP_QUERY = 2004  # Query a collection.
OP_GET_MORE = 2005  # Get more data from a query. See Cursors.
OP_DELETE = 2006  # Delete documents.
OP_KILL_CURSORS = 2007  # Notify database that the client has finished with the cursor.


def __pack_message(operation, data):
    """Takes message data and adds a message header based on the operation.

    Returns the resultant message string.
    """
    request_id = _randint()
    message = struct.pack("<i", 16 + len(data))
    message += struct.pack("<i", request_id)
    message += _ZERO_32
    message += struct.pack("<i", operation)
    return request_id, message + data


def query(options, collection_name, num_to_skip, num_to_return, request, field_selector=None):
    """Get a **query** message. """
    data = struct.pack("<I", options)
    data += _make_c_string(collection_name)
    data += struct.pack("<i", num_to_skip)
    data += struct.pack("<i", num_to_return)
    data += BSON.encode(request)
    if field_selector:
        data += BSON.encode(field_selector)
    return __pack_message(OP_QUERY, data)


def get_more(collection_name, num_to_return, cursor_id):
    """Get a **getMore** message."""
    data = _ZERO_32
    data += _make_c_string(collection_name)
    data += struct.pack("<i", num_to_return)
    data += struct.pack("<q", cursor_id)
    return __pack_message(OP_GET_MORE, data)


def insert(collection_name, docs, check_keys, safe, last_error_args):
    """Get an **insert** message. """
    data = _ZERO_32
    data += _make_c_string(collection_name)
    bson_data = "".join([BSON.encode(doc, check_keys) for doc in docs])
    if not bson_data:
        raise MokitoEmptyBulkError()
    data += bson_data
    if safe:
        (_, insert_message) = __pack_message(OP_INSERT, data)
        (request_id, error_message) = __last_error(last_error_args)
        return request_id, insert_message + error_message
    else:
        return __pack_message(OP_INSERT, data)


def update(collection_name, upsert, multi, spec, doc, safe, last_error_args):
    """Get an **update** message."""
    options = 0
    if upsert:
        options += 1
    if multi:
        options += 2

    data = _ZERO_32
    data += _make_c_string(collection_name)
    data += struct.pack("<i", options)
    data += BSON.encode(spec)
    data += BSON.encode(doc)
    if safe:
        (_, update_message) = __pack_message(OP_UPDATE, data)
        (request_id, error_message) = __last_error(last_error_args)
        return request_id, update_message + error_message
    else:
        return __pack_message(OP_UPDATE, data)


def delete(collection_name, spec, safe, last_error_args):
    """Get a **delete** message."""
    data = _ZERO_32
    data += _make_c_string(collection_name)
    data += _ZERO_32
    data += BSON.encode(spec)
    if safe:
        (_, remove_message) = __pack_message(OP_DELETE, data)
        (request_id, error_message) = __last_error(last_error_args)
        return request_id, remove_message + error_message
    else:
        return __pack_message(OP_DELETE, data)


def kill_cursors(*cursor_ids):
    """Get a **killCursors** message."""
    data = _ZERO_32
    data += struct.pack("<i", len(cursor_ids))
    for cursor_id in cursor_ids:
        data += struct.pack("<q", cursor_id)
    return __pack_message(OP_KILL_CURSORS, data)


def unpack_response(response):
    """Unpack a response from the database.

    Check the response for errors and unpack, returning a dictionary containing the response data.
    :Parameters:
      - `response`: byte string as returned from the database
      - `cursor_id`: cursor_id we sent to get this response - used for raising an informative
      exception when we get cursor id not valid at server response
    """
    response_flag = struct.unpack("<i", response[:4])[0]
    if response_flag & 1:
        # Shouldn't get this response if we aren't doing a getMore
        raise MokitoInvalidCursorError()

    elif response_flag & 2:
        error_object = BSON(response[20:]).decode()
        if error_object["$err"] == "not master":
            raise MokitoMasterChangedError()
        raise MokitoDriverError("MongoDB: %s" % error_object["$err"])

    result = {"cursor_id": struct.unpack("<q", response[4:12])[0],
              "starting_from": struct.unpack("<i", response[12:16])[0],
              "number_returned": struct.unpack("<i", response[16:20])[0],
              "data": decode_all(response[20:], DEFAULT_CODEC_OPTIONS)}

    assert len(result["data"]) == result["number_returned"]
    return result


def index_document(index_list):
    """Helper to generate an index specifying document.
    Takes a list of (key, direction) pairs.
    """
    if isinstance(index_list, dict):
        raise TypeError("passing a dict to sort/create_index/hint is not "
                        "allowed - use a list of tuples instead. did you "
                        "mean %r?" % list(index_list.iteritems()))
    elif not isinstance(index_list, list):
        raise TypeError("must use a list of (key, direction) pairs, not: " + repr(index_list))
    if not len(index_list):
        raise ValueError("key_or_list must not be the empty list")

    index = SON()
    for (key, value) in index_list:
        if not isinstance(key, basestring):
            raise TypeError("first item in each key pair must be a string")
        if value not in [ASCENDING, DESCENDING, GEO2D]:
            raise TypeError("second item in each key pair must be ASCENDING, DESCENDING, or GEO2D")
        index[key] = value
    return index

# def _password_digest(username, password):
#     """Get a password digest to use for authentication.
#     """
#     if not isinstance(password, basestring):
#         raise TypeError("password must be an instance of basestring")
#     if not isinstance(username, basestring):
#         raise TypeError("username must be an instance of basestring")
#
#     md5hash = hashlib.md5()
#     md5hash.update("%s:mongo:%s" % (username.encode('utf-8'),
#                                     password.encode('utf-8')))
#     return unicode(md5hash.hexdigest())
#
#
# def _auth_key(nonce, username, password):
#     """Get an auth key to use for authentication.
#     """
#     digest = _password_digest(username, password)
#     md5hash = hashlib.md5()
#     md5hash.update("%s%s%s" % (nonce, unicode(username), digest))
#     return unicode(md5hash.hexdigest())



# coding: utf-8

import socket
import struct
import urlparse
from functools import partial

from tornado.iostream import IOStream
from tornado.gen import coroutine, Return
from tornado.concurrent import Future
from tornado.ioloop import IOLoop

from errors import (
    MokitoResponseError,
    MokitoInvalidURIError,
    MokitoConnectionError
)
from message import unpack_response


class Connection(object):

    def __init__(self, uri):
        try:
            _uri = urlparse.urlparse(uri)
            self._host = _uri.hostname
            self._port = int(_uri.port or 27017)
        except:
            raise MokitoInvalidURIError()

        self.__stream = None

    def simple_connect(self):
        try:
            self.__stream = IOStream(socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0))
            self.__stream.connect((self._host, self._port))
        except socket.error as e:
            raise MokitoConnectionError(e)

    def close(self):
        self.__stream.close()

    @coroutine
    def send_message(self, request, safe=True):
        xx = Future()
        IOLoop.current().add_callback(partial(lambda x: x.set_result(None), xx))
        res = yield self._send_message(request, safe, xx)
        raise Return(res)

    @coroutine
    def _send_message(self, request, safe, future):
        yield future

        self.__stream.write(request[1])
        if safe:
            length = yield self._get_header(request[0])
            res = yield self._get_response(length - 16)
            raise Return(res)

    @coroutine
    def _get_header(self, request_id):
        header = yield self.__stream.read_bytes(16)
        length = int(struct.unpack("<i", header[:4])[0])
        _request_id = struct.unpack("<i", header[8:12])[0]
        assert request_id == _request_id, "ids don't match %r %r" % (request_id, _request_id)
        assert 1 == struct.unpack("<i", header[12:])[0]
        raise Return(length)

    @coroutine
    def _get_response(self, length):
        response = yield self.__stream.read_bytes(length)
        response = unpack_response(response)
        if response and response['data'] and response['data'][0].get('err') and \
                response['data'][0].get('code'):
            raise MokitoResponseError(response['data'][0]['err'], code=response['data'][0]['code'])
        raise Return(response)



# coding: utf-8

import unittest

from tornado.testing import gen_test, AsyncTestCase
from tornado.ioloop import IOLoop

import mokito
from tests.util import TEST_DB_NAME, TEST_DB_URI


# TODO: add test connection pool

class ConnectionTestCase(AsyncTestCase):

    def get_new_ioloop(self):
        return IOLoop.instance()

    def setUp(self):
        super(ConnectionTestCase, self).setUp()
        self.db = mokito.Client(TEST_DB_NAME, TEST_DB_URI)

    @gen_test
    def tearDown(self):
        yield self.db.command('dropDatabase')
        super(ConnectionTestCase, self).tearDown()

    def test_db_name(self):
        self.assertEqual(self.db._db_name, TEST_DB_NAME)

    @gen_test
    def test_command_1(self):
        res = yield self.db.command('buildinfo')
        self.assertIsNotNone(res)
        self.assertTrue(bool(res['ok']))

    @gen_test
    def test_command_2(self):
        yield self.db.foo.insert({'foo': 'bar'})

        res = yield self.db.foo.find_one()
        self.assertIsNotNone(res)

        res = yield self.db.command('drop', 'foo')
        self.assertIsNotNone(res)
        self.assertTrue(bool(res['ok']))

        res = yield self.db.foo.find_one()
        self.assertIsNone(res)

    @unittest.skip('not implemented method "close"')
    def test_close(self):
        self.db.close()

    @gen_test
    def test_uri_wrong(self):
        db = mokito.Client('wrong', "mongodb://127.0.0.1:27017aaa")
        with self.assertRaises(mokito.errors.MokitoInvalidURIError):
            yield db.command('buildinfo')

    @gen_test
    def test_collection_names(self):
        col = [u'foo', u'bar']
        for i in col+[u'tst1', u'tst2']:
            yield self.db.command('drop', i)

        for i in col:
            yield self.db[i].insert({'foo': 1})

        res = yield self.db.collection_names()
        self.assertListEqual(res, col)

        yield self.db.command('drop', 'bar')
        res = yield self.db.collection_names()
        self.assertListEqual(res, [u'foo'])

        yield self.db.command('drop', 'foo')
        res = yield self.db.collection_names()
        self.assertListEqual(res, [])



# coding: utf-8

import unittest

from mokito.fields import make_field, TupleField, ListField, NoneField


class TestNoneFields(unittest.TestCase):
    def test_list_field_1(self):
        f = make_field([None])
        self.assertIsInstance(f, ListField)
        self.assertIsInstance(f._rules, NoneField)
        f[1] = 123
        f.append('foo')
        self.assertIsNone(f.pop(0).value)
        self.assertListEqual(f.value, [123, 'foo'])
        f[1] = None
        self.assertListEqual(f.value, [123, None])
        del f[1]
        self.assertListEqual(f.value, [123])

        f.append(('bar', 2))
        self.assertListEqual(f.value, [123, ['bar', 2]])
        self.assertEqual(f[1][1].value, 2)
        self.assertEqual(f['1.1'].value, 2)

    def test_tuple_field_1(self):
        f = make_field((None, str))
        self.assertIsInstance(f, TupleField)
        self.assertListEqual(f.value, [None, None])
        f[0] = 123
        f[1] = 123
        self.assertListEqual(f.value, [123, '123'])
        f[0] = 'foo'
        self.assertListEqual(f.value, ['foo', '123'])
        del f[0]
        self.assertListEqual(f.value, [None, '123'])

    def test_tuple_field_2(self):
        f = make_field((str, [None]))
        self.assertListEqual(f.value, [None, []])
        f.set(['foo', [1, 'bar']])
        self.assertListEqual(f.value, ['foo', [1, 'bar']])
        f[1].append(123.4)
        self.assertListEqual(f.value, ['foo', [1, 'bar', 123.4]])
        del f['1.1']
        self.assertListEqual(f.value, ['foo', [1, 123.4]])
        f[1] = None
        self.assertListEqual(f.value, ['foo', []])

    def test_dict_field_1(self):
        f = make_field({
            'a': int,
            'b': None
        })
        self.assertDictEqual(f.value, {'a': None, 'b': None})

        f.set({'a': 123, 'b': 456})
        self.assertDictEqual(f.value, {'a': 123, 'b': 456})

        f.set({'a': 1, 'b': [1, 2, 3]})
        self.assertDictEqual(f.value, {'a': 1, 'b': [1, 2, 3]})

        f.clear()
        self.assertDictEqual(f.value, {'a': None, 'b': None})

    def test_list_query_1(self):
        f = make_field([None])
        self.assertDictEqual(f.query, {})
        f.append('foo')
        self.assertDictEqual(f.query, {'$set': ['foo']})
        f.append(123)
        self.assertDictEqual(f.query, {'$set': ['foo', 123]})
        del f[0]
        self.assertDictEqual(f.query, {'$set': [123]})
        f.dirty_clear()
        self.assertDictEqual(f.query, {})
        f[1] = ('x', 456.7)
        self.assertDictEqual(f.query, {'$set': [123, ['x', 456.7]]})
        f.pop(0)
        self.assertDictEqual(f.query, {'$set': [['x', 456.7]]})
        f.dirty_clear()
        del f[0]
        self.assertDictEqual(f.query, {'$set': []})

    def test_tuple_query_1(self):
        f = make_field((int, None))
        self.assertDictEqual(f.query, {})
        f[0] = 123
        self.assertDictEqual(f.query, {'$set': [123, None]})
        f[1] = 'foo'
        self.assertDictEqual(f.query, {'$set': [123, 'foo']})
        f.dirty_clear()
        self.assertDictEqual(f.query, {})
        del f[0]
        self.assertDictEqual(f.query, {'$set': [None, 'foo']})

    def test_list_tuple_query_1(self):
        f = make_field((str, [None]))
        self.assertDictEqual(f.query, {})
        f.set(['foo', [1]])
        self.assertDictEqual(f.query, {'$set': ['foo', [1]]})
        f['1.2'] = 'bar'
        self.assertDictEqual(f.query, {'$set': ['foo', [1, None, 'bar']]})

    def test_list_query_2(self):
        f = make_field([{'a': int, 'b': None}])
        self.assertDictEqual(f.query, {})
        f.append({'a': 1})
        self.assertDictEqual(f.query, {'$set': [{'a': 1, 'b': None}]})
        f.append({'b': 'foo'})
        self.assertDictEqual(f.query, {'$set': [{'a': 1, 'b': None}, {'a': None, 'b': 'foo'}]})
        f[0] = {'a': 123, 'b': 456.7}
        self.assertDictEqual(f.query, {'$set': [{'a': 123, 'b': 456.7}, {'a': None, 'b': 'foo'}]})



# coding: utf-8

from bson import ObjectId
from tornado.testing import gen_test, AsyncTestCase
from tornado.ioloop import IOLoop
from pymongo import ASCENDING

import mokito
from tests.util import TEST_DB_NAME, TEST_DB_URI


class CursorTestCase(AsyncTestCase):
    def get_new_ioloop(self):
        return IOLoop.instance()

    def setUp(self):
        super(CursorTestCase, self).setUp()
        self.db = mokito.Client(TEST_DB_NAME, TEST_DB_URI)

    @gen_test
    def tearDown(self):
        yield self.db.command('dropDatabase')
        super(CursorTestCase, self).tearDown()

    def test_collection_name(self):
        cur = self.db.foo
        self.assertEqual(cur.collection_name, 'foo')

    def test_full_collection_name(self):
        def cn(name=None):
            return '%s.%s' % (TEST_DB_NAME, name or 'foo')

        cur = self.db.foo
        self.assertEqual(cur.full_collection_name, cn())
        self.assertEqual(cur._full_collection_name('$cmd'), cn('$cmd'))

    @gen_test
    def test_insert_1(self):
        res = yield self.db.foo.insert({'foo': 1})
        self.assertIsInstance(res, ObjectId)

    @gen_test
    def test_insert_2(self):
        res = yield self.db.foo.insert([{'bar': 1}, {'bar': 2}])
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], ObjectId)
        self.assertIsInstance(res[1], ObjectId)

    @gen_test
    def test_insert_3(self):
        _id = ObjectId()
        res = yield self.db.foo.insert({'foo': 3, '_id': _id})
        self.assertEqual(_id, res)

    @gen_test
    def test_insert_4(self):
        _id = ObjectId()
        data = {'_id': _id, 'foo': 1}
        res = yield self.db.foo.insert(data, False)
        self.assertEqual(_id, res)

        res = yield self.db.foo.find_one()
        self.assertDictEqual(data, res)

    @gen_test
    def test_insert_wrong(self):
        _id = ObjectId()
        data = {'foo': 1, '_id': _id}
        yield self.db.foo.insert(data)
        with self.assertRaises(mokito.errors.MokitoResponseError):
            yield self.db.foo.insert(data)

    @gen_test
    def test_count_1(self):
        res = yield self.db.foo.count()
        self.assertFalse(res)

    @gen_test
    def test_count_2(self):
        yield self.db.foo.insert([{'bar': 1}, {'bar': 2}])
        res = yield self.db.foo.count()
        self.assertEqual(res, 2)

        res = yield self.db.foo.count({'bar': 1})
        self.assertEqual(res, 1)

    @gen_test
    def test_remove_1(self):
        _id1 = ObjectId()
        data = [{'bar': 1, '_id': _id1}, {'bar': 2}, {'bar': 3}]

        yield self.db.foo.insert(data)
        res = yield self.db.foo.remove()
        self.assertEqual(res, 3)
        res = yield self.db.foo.count()
        self.assertFalse(res)

        yield self.db.foo.insert(data)
        res = yield self.db.foo.remove(_id1)
        self.assertEqual(res, 1)
        res = yield self.db.foo.count()
        self.assertEqual(res, 2)

    @gen_test
    def test_remove_2(self):
        _id1 = ObjectId()
        _id2 = ObjectId()

        yield self.db.foo.insert([{'bar': 1, '_id': _id1}, {'bar': 2}])
        res = yield self.db.foo.remove(_id2)
        self.assertEqual(res, 0)
        res = yield self.db.foo.count()
        self.assertEqual(res, 2)

    @gen_test
    def test_remove_3(self):
        yield self.db.foo.insert([{'bar': 1}, {'bar': 1}, {'bar': 2}])
        res = yield self.db.foo.remove({'bar': 1})
        self.assertEqual(res, 2)
        res = yield self.db.foo.count()
        self.assertEqual(res, 1)

    @gen_test
    def test_find_one(self):
        res = yield self.db.foo.find_one()
        self.assertIsNone(res)

        _id1 = ObjectId()
        data1 = {'_id': _id1, 'bar': 1, 'foo': 'foo'}
        _id2 = ObjectId()
        data2 = {'_id': _id2, 'bar': 2, 'foo': 'foo'}
        yield self.db.foo.insert([data1, data2])

        res = yield self.db.foo.find_one(_id1)
        self.assertDictEqual(data1, res)

        res = yield self.db.foo.find_one({'bar': 2})
        self.assertDictEqual(data2, res)

        res = yield self.db.foo.find_one({'bar': 3})
        self.assertIsNone(res)

        res = yield self.db.foo.find_one(_id1, {'bar': 1, 'foo': 1})
        self.assertDictEqual(data1, res)

        res = yield self.db.foo.find_one(_id1, {'bar': 1})
        self.assertDictEqual({'_id': _id1, 'bar': 1}, res)

        res = yield self.db.foo.find_one(_id2, ['foo'])
        self.assertDictEqual({'_id': _id2, 'foo': 'foo'}, res)

    @gen_test
    def test_find_1(self):
        res = yield self.db.foo.find()
        self.assertIsInstance(res, list)
        self.assertFalse(res)

        _id1 = ObjectId()
        data1 = {'_id': _id1, 'bar': 1, 'foo': 'foo'}
        _id2 = ObjectId()
        data2 = {'_id': _id2, 'bar': 2, 'foo': 'foo'}
        yield self.db.foo.insert([data1, data2])

        res = yield self.db.foo.find()
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 2)

        res = yield self.db.foo.find({'_id': _id1})
        self.assertDictEqual(data1, res[0])

        res = yield self.db.foo.find({'bar': 2})
        self.assertDictEqual(data2, res[0])

        res = yield self.db.foo.find({'foo': 'foo'}, ['bar'], limit=1)
        self.assertDictEqual({'_id': _id1, 'bar': 1}, res[0])

        res = yield self.db.foo.find({'foo': 'foo'}, ['bar'], skip=1)
        self.assertDictEqual({'_id': _id2, 'bar': 2}, res[0])

    @gen_test
    def test_find_2(self):
        _id1 = ObjectId()
        data1 = {'_id': _id1, 'bar': 1, 'foo': 'foo'}
        _id2 = ObjectId()
        data2 = {'_id': _id2, 'bar': 2, 'foo': 'foo'}
        yield self.db.foo.insert([data1, data2])

        res = yield self.db.foo.find(sort=[('bar', ASCENDING)])
        self.assertDictEqual(data1, res[0])
        self.assertDictEqual(data2, res[1])

        res = yield self.db.foo.find(sort='-bar')
        self.assertDictEqual(data2, res[0])
        self.assertDictEqual(data1, res[1])

    @gen_test
    def test_update_1(self):
        _id1 = ObjectId()
        data1 = {'_id': _id1, 'bar': 1, 'foo': 'foo'}
        _id2 = ObjectId()
        data2 = {'_id': _id2, 'bar': 2, 'foo': 'foo'}
        yield self.db.foo.insert([data1, data2])

        res = yield self.db.foo.update({'_id': _id1}, {'foo': 'baz'}, no_replace=True)
        self.assertTrue(res)
        res = yield self.db.foo.find()
        self.assertDictEqual({'_id': _id1, 'bar': 1, 'foo': 'baz'}, res[0])
        self.assertDictEqual(data2, res[1])

        res = yield self.db.foo.update({'_id': _id2}, {'foo': 'baz'})
        self.assertTrue(res)
        res = yield self.db.foo.find()
        self.assertDictEqual({'_id': _id1, 'bar': 1, 'foo': 'baz'}, res[0])
        self.assertDictEqual({'_id': _id2, 'foo': 'baz'}, res[1])

    @gen_test
    def test_update_2(self):
        _id1 = ObjectId()
        data1 = {'_id': _id1, 'bar': 1, 'foo': 'foo'}
        _id2 = ObjectId()
        data2 = {'_id': _id2, 'bar': 2, 'foo': 'foo'}
        yield self.db.foo.insert([data1, data2])

        res = yield self.db.foo.update({'foo': 'foo'}, {"$set": {'bar': 3}})
        self.assertTrue(res)
        res = yield self.db.foo.find()
        self.assertDictEqual({'_id': _id1, 'bar': 3, 'foo': 'foo'}, res[0])
        self.assertDictEqual(data2, res[1])

        res = yield self.db.foo.update({'foo': 'foo'}, {"$set": {'bar': 4}}, multi=True)
        self.assertTrue(res)
        res = yield self.db.foo.find()
        self.assertDictEqual({'_id': _id1, 'bar': 4, 'foo': 'foo'}, res[0])
        self.assertDictEqual({'_id': _id2, 'bar': 4, 'foo': 'foo'}, res[1])

    @gen_test
    def test_update_3(self):
        _id1 = ObjectId()
        data1 = {'_id': _id1, 'bar': 1, 'foo': 'foo'}
        _id2 = ObjectId()
        data2 = {'_id': _id2, 'bar': 2, 'foo': 'foo'}
        yield self.db.foo.insert(data1)

        res = yield self.db.foo.update({'_id': _id2}, data2)
        self.assertFalse(res)
        res = yield self.db.foo.find_one(_id2)
        self.assertIsNone(res)

        res = yield self.db.foo.update({'_id': _id2}, data2, upsert=True)
        self.assertEqual(_id2, res)
        res = yield self.db.foo.find_one(_id2)
        self.assertDictEqual(data2, res)

    @gen_test
    def test_distinct_1(self):
        _id1 = ObjectId()
        data1 = {'_id': _id1, 'bar': 1, 'foo': 'foo'}
        _id2 = ObjectId()
        data2 = {'_id': _id2, 'bar': 2, 'foo': 'foo'}
        yield self.db.foo.insert([data1, data2])

        res = yield self.db.foo.distinct('bar')
        self.assertListEqual(res, [1, 2])

    @gen_test
    def test_distinct_2(self):
        _id1 = ObjectId()
        data1 = {'_id': _id1, 'bar': 1, 'foo': 'foo'}
        _id2 = ObjectId()
        data2 = {'_id': _id2, 'bar': 2, 'foo': 'bar'}
        yield self.db.foo.insert([data1, data2])

        res = yield self.db.foo.distinct('bar', {'foo': 'bar'})
        self.assertListEqual(res, [2])



# coding: utf-8

from UserList import UserList

from bson import DBRef, ObjectId
from tornado.gen import coroutine, Return

from manage import ModelManager
from fields import DictField
from errors import MokitoORMError

from tools import SEPARATOR

DEFAULT_URI = "mongodb://127.0.0.1:27017"
DEFAULT_CONNECTIONS = 10


class ModelMeta(type):
    def __new__(mcs, name, bases, attr):
        scheme = attr.get('scheme')
        if scheme is not None and not isinstance(scheme, dict):
            attr['scheme'] = getattr(scheme, 'scheme')

        _cls = type.__new__(mcs, name, bases, attr)
        if name != 'Model' and name != 'Document':
            ModelManager.add(_cls)
        return _cls


class Model(DictField):
    __metaclass__ = ModelMeta
    scheme = {}

    def __init__(self, data=None, **kwargs):
        super(Model, self).__init__(self.scheme)
        if data:
            self.set(data, **kwargs)

    def get(self, key=None, **kwargs):
        if isinstance(key, (list, tuple)):
            data = {}
            fields = []
            for i in key:
                if hasattr(self, i):
                    data[i] = getattr(self, i)
                else:
                    fields.append(i)
            return dict(super(Model, self).get(fields, **kwargs), **data)

        return super(Model, self).get(key, **kwargs)


class Documents(UserList):
    def __repr__(self):
        return '<%s: [%s]>' % (self.__class__.__name__, ', '.join(map(str, self.data)))

    def __getitem__(self, key):
        k1, _, k2 = str(key).partition(SEPARATOR)
        ret = self.data[int(k1)]
        return ret[k2] if k2 else ret

    def __call__(self, method_name, *args, **kwargs):
        return [getattr(i, method_name)(*args, **kwargs) for i in self.data]

    def dirty_clear(self):
        map(lambda i: i.dirty_clear(), self.data)

    @coroutine
    def reread(self, *fields):
        # TODO: добавить кэш
        yield [i.reread(*fields) for i in self.data]

    @coroutine
    def remove(self, safe=True):
        if self.data:
            ids = [i._id for i in self.data]
            cur = self.data[0].get_cursor()
            yield cur.remove({"_id": {"$in": ids}}, safe)
            for i in self.data:
                i._id = None

    @coroutine
    def save(self, safe=True):
        yield [i.save(safe) for i in self.data]

    def filter(self, fn):
        return Documents(filter(fn, self.data))

    def get(self, key, **kwargs):
        return [i.get(key, **kwargs) for i in self.data]


class Document(Model):
    __uri__ = DEFAULT_URI
    __database__ = None
    __collection__ = None
    __connect_count__ = DEFAULT_CONNECTIONS
    BY_ID_ERROR = 'Запись не найдена'

    def __init__(self, data=None, **kwargs):
        self.loaded = False
        self._id = None
        if data:
            _id = data.pop('_id', None)
            if _id:
                self._id = ObjectId(_id)
        super(Document, self).__init__(data, **kwargs)

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__, self._id)

    def __call__(self, method_name, *args, **kwargs):
        return getattr(self, method_name)(*args, **kwargs)

    @property
    def self_value(self):
        return self.get_dbref() if self._id else None

    @property
    def id(self):
        if self._id is not None:
            return str(self._id)

    def get_dbref(self):
        return DBRef(self.__collection__, self._id)

    def set_dbref(self, dbref):
        self.__collection__ = dbref.collection
        self._id = dbref.id

    dbref = property(get_dbref, set_dbref)

    def clear(self):
        super(Document, self).clear()
        self._id = None

    def set(self, value, **kwargs):
        if isinstance(value, DBRef):
            self.dbref = value
        else:
            value = value.copy()
            _id = value.pop('_id', None)
            if _id is not None:
                self._id = ObjectId(_id)
            super(Document, self).set(value, **kwargs)

    @classmethod
    def get_cursor(cls, database=None, collection=None):
        db = database or cls.__database__
        try:
            return ModelManager.databases[db][collection or cls.__collection__]
        except KeyError:
            raise MokitoORMError('Connection to the database "%s" not found' % db)

    @coroutine
    def reread(self, *fields):
        """
        Read the object again.
        """
        if fields:
            yield [self[i].reread() for i in fields]

        elif self._id:
            cur = self.get_cursor()
            data = yield cur.find_one(self._id, fields=self.scheme.keys())
            if data:
                self.clear()
                self.set(data, inner=True)
                self.dirty_clear()
                self.loaded = True
                raise Return(True)

    @classmethod
    @coroutine
    def by_id(cls, _id):
        if _id:
            try:
                self = yield cls.find_one(_id)
            except:
                self = None
            if self:
                raise Return(self)
        raise MokitoORMError(cls.BY_ID_ERROR)

    @classmethod
    @coroutine
    def find_one_raw(cls, spec_or_id, *fields):
        cur = cls.get_cursor()
        data = yield cur.find_one(spec_or_id, fields=fields)
        raise Return(data)

    @classmethod
    @coroutine
    def find_one(cls, spec_or_id):
        cur = cls.get_cursor()
        fields = cls.scheme.keys()
        data = yield cur.find_one(spec_or_id, fields=fields)
        if data:
            self = cls(data, inner=True)
            self.dirty_clear()
            self.loaded = True
            raise Return(self)

    @classmethod
    @coroutine
    def find_raw(cls, spec, *fields, **kwargs):
        cur = cls.get_cursor()
        kw = {
            'fields': fields,
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 0),
            'sort': kwargs.get('sort'),
            'hint': kwargs.get('hint')
        }
        data = yield cur.find(spec, **kw)
        raise Return(data)

    @classmethod
    @coroutine
    def find(cls, spec=None, skip=0, limit=0, sort=None, hint=None):
        cur = cls.get_cursor()
        fields = cls.scheme.keys()
        data = yield cur.find(spec, fields, skip, limit, sort=sort, hint=hint)
        res = Documents(cls(i, inner=True) for i in data)
        res.dirty_clear()
        raise Return(res)

    @classmethod
    @coroutine
    def count(cls, spec=None):
        cur = cls.get_cursor()
        data = yield cur.count(spec)
        raise Return(data)

    @classmethod
    @coroutine
    def distinct(cls, key, spec=None):
        cur = cls.get_cursor()
        data = yield cur.distinct(key, spec)
        raise Return(data)

    @coroutine
    def save(self, safe=True):
        if self._id is None:
            cur = self.get_cursor()
            self._id = yield cur.insert(super(Document, self).self_value, safe=safe)
            self.dirty_clear()
            if safe:
                raise Return(True)

        elif self.dirty:
            cur = self.get_cursor()
            res = yield cur.update(self._id, self.query, safe=safe)
            if res:
                self.dirty_clear()
                if safe:
                    raise Return(res)

    @coroutine
    def remove(self, safe=True):
        cur = self.get_cursor()
        yield cur.remove(self._id, safe)
        self._id = None



# coding: utf-8

import unittest

from mokito.fields import (
    make_field,
    IntField,
    FloatField,
    StringField,
    DictField,
    TupleField,
    ListField
)


class TestCollectionFieldsDefault(unittest.TestCase):
    def test_list_field_1(self):
        f = make_field([123])
        self.assertIsInstance(f, ListField)
        self.assertIsInstance(f._rules, IntField)
        self.assertListEqual(f.value, [123])

        f.append('111')
        self.assertListEqual(f.value, [123, 111])

        f.set(None)
        self.assertListEqual(f.value, [])

    def test_tuple_field_1(self):
        f = make_field((123, 'foo'))
        self.assertIsInstance(f, TupleField)
        self.assertIsInstance(f._val['0'], IntField)
        self.assertTrue(f == f._val['0']._parent)
        self.assertIsInstance(f._val['1'], StringField)
        self.assertTrue(f == f._val['1']._parent)
        self.assertListEqual(f.value, [123, 'foo'])

        f[0] = 456
        self.assertEqual(f.value, [456, 'foo'])
        self.assertEqual(f.self_value, [456, 'foo'])
        self.assertEqual(f.get(), [456, 'foo'])
        self.assertTrue(f == f[0]._parent)

        f[1] = 456
        self.assertEqual(f.value, [456, '456'])
        self.assertEqual(f.self_value, [456, '456'])
        self.assertEqual(f.get(), [456, '456'])
        self.assertTrue(f == f[1]._parent)

        f[0] = None
        self.assertEqual(f.value, [None, '456'])
        self.assertEqual(f.self_value, [None, '456'])
        self.assertEqual(f.get(), [None, '456'])
        self.assertTrue(f == f[0]._parent)

        del f[1]
        self.assertEqual(f.value, [None, None])
        self.assertEqual(f.self_value, [None, None])
        self.assertEqual(f.get(), [None, None])
        self.assertTrue(f == f[1]._parent)

    def test_dict_field_1(self):
        data = {
            'f0': 'foo',
            'f1': 123,
            'f2': 456.7,
        }
        f = make_field(data)
        self.assertIsInstance(f, DictField)
        self.assertIsInstance(f._val['f0'], StringField)
        self.assertTrue(f == f._val['f0']._parent)
        self.assertIsInstance(f._val['f1'], IntField)
        self.assertTrue(f == f._val['f1']._parent)
        self.assertIsInstance(f._val['f2'], FloatField)
        self.assertTrue(f == f._val['f2']._parent)

        self.assertDictEqual(f.value, {'f0': 'foo', 'f1': 123, 'f2': 456.7})
        f.set({'f0': 'bar', 'f1': 456})
        self.assertDictEqual(f.value, {'f0': 'bar', 'f1': 456, 'f2': 456.7})
        f['f2'] = 9
        self.assertDictEqual(f.value, {'f0': 'bar', 'f1': 456, 'f2': 9.0})

    def test_list_query_1(self):
        f = make_field(['foo'])
        self.assertDictEqual(f.query, {'$set': ['foo']})
        f.append('bar')
        self.assertDictEqual(f.query, {'$set': ['foo', 'bar']})
        del f[0]
        self.assertDictEqual(f.query, {'$set': ['bar']})
        f.dirty_clear()
        self.assertDictEqual(f.query, {})
        f[1] = 'x'
        self.assertDictEqual(f.query, {'$set': ['bar', 'x']})

    def test_tuple_query_1(self):
        f = make_field((123, str))
        self.assertDictEqual(f.query, {'$set': [123, None]})
        f[1] = 'foo'
        self.assertDictEqual(f.query, {'$set': [123, 'foo']})
        f.dirty_clear()
        self.assertDictEqual(f.query, {})
        del f[0]
        self.assertDictEqual(f.query, {'$set': [None, 'foo']})



# coding: utf-8

from client import Client


class ModelManager(object):
    models = {}
    databases = {}

    @classmethod
    def add(cls, _class):
        cls.models[_class.__name__] = _class

        db_name = getattr(_class, '__database__', None)
        if db_name and db_name not in cls.databases:
            cls.databases[db_name] = Client(db_name,
                                            getattr(_class, '__uri__', None),
                                            getattr(_class, '__connect_count__', None))

    @classmethod
    def close_db(cls):
        for i in cls.databases.values():
            i.close()
        # cls.databases = {}




# coding: utf-8


class MokitoError(Exception):
    """Base class for all exceptions. """


class MokitoDriverError(MokitoError):
    pass


class MokitoEmptyBulkError(MokitoDriverError):
    message = "cannot do an empty bulk insert"


class MokitoMasterChangedError(MokitoDriverError):
    message = "master has changed"


class MokitoInvalidURIError(MokitoDriverError):
    message = "Invalid URI"


class MokitoInvalidCursorError(MokitoDriverError):
    message = "Cursor not valid at server"


class MokitoParamError(MokitoDriverError):
    message = "Invalid parameters"


class MokitoConnectionError(MokitoDriverError):
    pass


class MokitoResponseError(MokitoDriverError):

    def __init__(self, msg, code=None):
        self.code = code
        self.msg = msg

    def __unicode__(self):
        return u'Response error: %s code:%s' % (self.msg, self.code or '')

    def __str__(self):
        return str(self.__unicode__())


class MokitoORMError(MokitoError):
    pass


class MokitoDBREFError(MokitoORMError):
    pass


class MokitoChoiceError(MokitoORMError):
    pass



# coding: utf-8

from bson import ObjectId
from pymongo import ASCENDING, DESCENDING


SEPARATOR = '.'


def fields2sort(sort):
    if isinstance(sort, basestring):
        sort = [sort]
    return [(i[1:], DESCENDING,) if i[0] == '-' else
            (i, ASCENDING,) if isinstance(i, basestring) else
            i for i in sort]


def norm_spec(spec):
    if spec is None:
        return {}

    if isinstance(spec, basestring):
        return {"_id": ObjectId(spec)}

    if isinstance(spec, ObjectId):
        return {"_id": spec}

    return spec
